"""
    Transformations that rewrite reified constraints as needed.

    There are three types of reification (BV=BoolVar, BE=BoolExpr):

    =============  ============================================================
    ``BV -> BE``   single implication, from var to expression                 
    ``BV <- BE``   single implication, from expression to var                
    ``BE == BV``   full reification / double implication (e.g. ``BV <-> BE``) 
    =============  ============================================================

    Using logical operations, they can be decomposed and rewritten to each other.

    This file implements:

    ==========================  =================================================================
    :func:`only_bv_reifies()`   transforms all reifications to ``BV -> BE`` or ``BV == BE``      
    :func:`only_implies()`      transforms all reifications to ``BV -> BE`` form                 
    :func:`reify_rewrite()`     rewrites reifications not supported by a solver to ones that are 
    ==========================  =================================================================
"""
import copy
from ..expressions.core import Operator, Comparison, Expression
from ..expressions.globalconstraints import GlobalConstraint
from ..expressions.globalfunctions import Element
from ..expressions.variables import _BoolVarImpl, _NumVarImpl
from ..expressions.python_builtins import all
from ..expressions.utils import is_any_list
from .flatten_model import flatten_constraint, get_or_make_var
from .negation import recurse_negation

def only_bv_reifies(constraints, csemap=None):

    newcons = []
    for cpm_expr in constraints:
        if cpm_expr.name in ['->', "=="]:
            a0, a1 = cpm_expr.args
            if not isinstance(a0, _BoolVarImpl) and \
                    isinstance(a1, _BoolVarImpl):
                # BE -> BV :: ~BV -> ~BE
                if cpm_expr.name == '->':
                    newexpr = (~a1).implies(recurse_negation(a0))
                    newexpr = only_bv_reifies(flatten_constraint(newexpr, csemap=csemap), csemap=csemap)
                else:
                    newexpr = [a1 == a0]  # BE == BV :: BV == BE
                    if not a0.is_bool():
                        newexpr = flatten_constraint(newexpr, csemap=csemap)
                newcons.extend(newexpr)
            else:
                newcons.append(cpm_expr)
        else:
            newcons.append(cpm_expr)
    return newcons

def only_implies(constraints, csemap=None):
    """
        Transforms all reifications to ``BV -> BE`` form

        More specifically:

        .. code-block:: text

            BV0 -> BV2 == BV3 :: BV0 -> (BV2->BV3 & BV3->BV2)
                              :: BV0 -> (BV2->BV3) & BV0 -> (BV3->BV2)
                              :: BV0 -> (~BV2|BV3) & BV0 -> (~BV3|BV2)
            BV == BE :: ~BV -> ~BE, BV -> BE

        Assumes all constraints are in 'flat normal form' and all reifications have a variable in lhs. Hence, only apply
        AFTER :func:`~cpmpy.transformations.flatten_model.flatten_constraint()` and :func:`only_bv_reifies()`.
    """
    newcons = []
    retransform = []

    for cpm_expr in constraints:
        # Operators: check BE -> BV
        if cpm_expr.name == '->' and cpm_expr.args[1].name == '==':
            a0,a1 = cpm_expr.args
            if a1.args[0].is_bool() and a1.args[1].is_bool():
                # BV0 -> BV2 == BV3 :: BV0 -> (BV2->BV3 & BV3->BV2)
                #                   :: BV0 -> (BV2->BV3) & BV0 -> (BV3->BV2)
                #                   :: BV0 -> (~BV2|BV3) & BV0 -> (~BV3|BV2)
                bv2,bv3 = a1.args
                retransform.extend(( a0.implies(~bv2|bv3), a0.implies(~bv3|bv2) ))
            else:
                newcons.append(cpm_expr)

        # Comparisons: transform BV == BE
        elif cpm_expr.name == '==' and cpm_expr.args[0].is_bool():
            # a0 is a boolvar, because of previous transformation only_bv_reifies.
            a0,a1 = cpm_expr.args
            if isinstance(a1, _BoolVarImpl):
                # BVar0 == BVar1 special case, no need to re-transform
                newcons.append(a0.implies(a1))
                newcons.append(a1.implies(a0))
            elif not a1.is_bool():
                # if a rhs integer is involved with a lhs bool,
                # then it is actually an integer expression, keep
                newcons.append(cpm_expr)
            else:
                # BVar1 == BE0 :: BVar1 -> BE0, ~BVar1 -> ~BE0
                # assume that if a0 == a1 was fine, that a0 -> a1 is too
                # EXCEPT, optimisation a0 -> a1_0 & ... & a1_n :: a0 -> a1_0 & ... & a0 -> a1_n
                if a1.name == 'and':
                    # optimisation without going through retransform
                    newcons.extend(a0.implies(a1_i) for a1_i in a1.args)
                elif a1.has_subexpr():
                    # requires retransform
                    retransform.append(a0.implies(a1))
                else:
                    newcons.append(a0.implies(a1))

                neg_a1 = recurse_negation(a1)
                if neg_a1.name == 'and':
                    # optimisation without going through retransform
                    newcons.extend((~a0).implies(na1_i) for na1_i in neg_a1.args)
                elif neg_a1.has_subexpr():
                    # requires retransform
                    retransform.append((~a0).implies(neg_a1))
                else:
                    newcons.append((~a0).implies(neg_a1))
        else:
            # all other flat normal form expressions are fine
            newcons.append(cpm_expr)
    
    if len(retransform) != 0:
        newcons.extend(only_implies(only_bv_reifies(flatten_constraint(retransform, csemap=csemap), csemap=csemap), csemap=csemap))

    return newcons


def reify_rewrite(constraints, supported=frozenset(), csemap=None):
    """
        Rewrites reified constraints not natively supported by a solver,
        to a version that uses standard constraints and reification over equalities between variables.

        Input is expected to be in Flat Normal Form without unsupported globals present.
        (so after :func:`~cpmpy.transformations.flatten_model.flatten_constraint()` and :func:`~cpmpy.transformations.decompose_global.decompose_global()`)
        Output will also be in Flat Normal Form

        Boolean expressions ``and``, ``or``, and ``->`` and comparison expression ``IV1==IV2`` are assumed to support reification
        (actually currently all comparisons <op> in {'==', '!=', '<=', '<', '>=', '>'},
        ``IV1 <op> IV2`` are assumed to support reification ``BV -> (IV1 <op> IV2)``)

        :param supported: a (frozen)set of expression names that support reification in the solver, including
                          supported 'Left Hand Side (LHS)' expressions in reified comparisons, e.g. ``BV -> (LHS == V)``
    """
    if not is_any_list(constraints):
        # assume list, so make list
        constraints = [constraints]

    newcons = []
    for cpm_expr in constraints:
        assert isinstance(cpm_expr, Expression), f"Expected CPMpy Expression but got {cpm_expr}, " \
                                                 f"run transformations.normalize.make_cpm_expr first!"
        # check if reif, get (the index of) the Boolean subexpression BE
        boolexpr_index = None
        if cpm_expr.name == '->':
            if not isinstance(cpm_expr.args[0], _BoolVarImpl):  # BE -> BV
                boolexpr_index = 0
            elif not isinstance(cpm_expr.args[1], _BoolVarImpl):  # BV -> BE
                boolexpr_index = 1
            else:
                pass  # both BV
        elif cpm_expr.name == '==' and \
                cpm_expr.args[0].is_bool() and \
                not isinstance(cpm_expr.args[0], _BoolVarImpl) and \
                isinstance(cpm_expr.args[1], _BoolVarImpl):  # BE == BV
            boolexpr_index = 0

        if boolexpr_index is None:  # non-reified or variable-only reification
            newcons.append(cpm_expr)
        else:  # reification, check for rewrite
            boolexpr = cpm_expr.args[boolexpr_index]
            if isinstance(boolexpr, Operator):
                # Case 1, BE is Operator (and, or, ->)
                #   assume supported, return as is
                newcons.append(cpm_expr)
                #   could actually rewrite into list of clauses like to_cnf() does... not for here
            elif isinstance(boolexpr, GlobalConstraint):
                # Case 2, BE is a GlobalConstraint
                # replace BE by its decomposition, then flatten
                if boolexpr.name in supported:
                    newcons.append(cpm_expr)
                else:
                    raise ValueError(f"Unsupported boolexpr {boolexpr} in reification, run a suitable decomposition "
                                     f"transformation from `cpmpy.transformations.decompose_global` to decompose "
                                     f"unsupported global constraints")
            elif isinstance(boolexpr, Comparison):
                # Case 3, BE is Comparison(OP, LHS, RHS)
                op, (lhs, rhs) = boolexpr.name, boolexpr.args
                #   have list of supported lhs's such as sum and wsum...
                #   at the very least, (iv1 == iv2) == bv has to be supported
                if isinstance(lhs, _NumVarImpl) or lhs.name in supported:
                    newcons.append(cpm_expr)
                else:  # other cases, LHS is a total function
                    #     introduce aux var and bring function to toplevel
                    #     (AUX,c) = get_or_make_var(LHS)
                    #     return c+[Comp(OP,AUX,RHS) == BV] or +[Comp(OP,AUX,RHS) -> BV] or +[Comp(OP,AUX,RHS) <- BV]
                    (auxvar, cons) = get_or_make_var(lhs, csemap=csemap)
                    newcons += cons
                    reifexpr = copy.copy(cpm_expr)
                    reifexpr.args[boolexpr_index] = Comparison(op, auxvar, rhs)  # Comp(OP,AUX,RHS)
                    newcons.append(reifexpr)
            else:
                # don't think this will be reached
                newcons.append(cpm_expr)

    return newcons



"""
  Transforms non-equality comparisons into equality comparisons as needed.
  
  Let <op> be one of `==` or `!=`, `<`, `<=`, `>`, `>=`. Numeric expressions in **Flat Normal Form** are of the kind:

    - `NumExpr <op> IV`
    - `BoolVar == NumExpr <op> IV`
    - `BoolVar -> NumExpr <op> IV`
    - `NumExpr <op> IV -> BoolVar`

  The `NumExpr` can be a sum, wsum or global function with a non-bool return type.
    
  This file implements:
    - :func:`only_numexpr_equality()`:    transforms `NumExpr <op> IV` (also reified) to `(NumExpr == A) & (A <op> IV)` if not supported
"""

import copy
from .flatten_model import get_or_make_var
from ..expressions.core import Comparison, Operator
from ..expressions.utils import is_boolexpr
from ..expressions.variables import _NumVarImpl, _BoolVarImpl

def only_numexpr_equality(constraints, supported=frozenset(), csemap=None):
    """
        Transforms ``NumExpr <op> IV`` to ``(NumExpr == A) & (A <op> IV)`` if not supported.
        Also for the reified uses of `NumExpr`

        :param supported:  a (frozen)set of expression names that supports all comparisons in the solver
    """

    newlist = []
    for cpm_expr in constraints:

        if isinstance(cpm_expr, Operator) and cpm_expr.name == "->":
            cond, subexpr = cpm_expr.args
            if not isinstance(cond, _BoolVarImpl): # expr -> bv
                idx = 0
            elif not isinstance(subexpr, _BoolVarImpl): # bv -> expr
                idx = 1
            else: # bv -> bv
                newlist.append(cpm_expr)
                continue

            new_arg, new_cons = _rewrite_comparison(cpm_expr.args[idx], supported=supported,csemap=csemap)
            if new_arg is not cpm_expr.args[idx]: # changed
                cpm_expr = copy.copy(cpm_expr) # shallow copy
                cpm_expr.args[idx] = new_arg                
                cpm_expr.update_args(cpm_expr.args) # XXX redundant? we know it's flat so no subexprs anyway
            
            newlist += [cpm_expr] + new_cons

            
        elif isinstance(cpm_expr, Comparison):

            lhs, rhs = cpm_expr.args
            if cpm_expr.name == "==" and is_boolexpr(lhs) and is_boolexpr(rhs): # reification
                if not isinstance(lhs, _BoolVarImpl):  # expr == bv
                    idx = 0
                elif not isinstance(rhs, _BoolVarImpl):  # bv == expr
                    idx = 1
                else: # bv == bv
                    newlist.append(cpm_expr)
                    continue

                # identical to the above, but keep for readability?
                new_arg, new_cons = _rewrite_comparison(cpm_expr.args[idx], supported=supported,csemap=csemap)
                if new_arg is not cpm_expr.args[idx]: # changed
                    cpm_expr = copy.copy(cpm_expr) # shallow copy
                    cpm_expr.args[idx] = new_arg
                    cpm_expr.update_args(cpm_expr.args) # XXX redundant? we know it's flat so no subexprs anyway

                newlist += [cpm_expr] + new_cons

            elif cpm_expr.name != "==": # numerical comparison
                new_expr, new_cons = _rewrite_comparison(cpm_expr, supported=supported,csemap=csemap)
                newlist += [new_expr] + new_cons
            
            else:
                newlist.append(cpm_expr) # equality constraint, keep

        else:
            # default, keep original
            newlist.append(cpm_expr)
                
    return newlist


def _rewrite_comparison(cpm_expr, supported=frozenset(), csemap=None):
    """
    Rewrite a comparison to an equality comparison, and a defining constraint.

    E.g., max(x,y,z) < p is rewritten to:
        max(x,y,z) == iv & iv < p

    :param cpm_expr: the comparison to rewrite
    :param csemap: the cse map to use
    :return: the rewritten comparison and the defining constraint
    """
    if not isinstance(cpm_expr, Comparison):
        return cpm_expr, []

    lhs, rhs = cpm_expr.args # flat, so expression will be on left hand side
    if cpm_expr.name != "==" and not isinstance(lhs, _NumVarImpl) and lhs.name not in supported:
        # lhs is unsupported, rewrite to `(LHS == A) & (A <op> RHS)`
        cpm_expr = copy.copy(cpm_expr)
        new_lhs, new_cons = get_or_make_var(lhs, csemap=csemap)
        cpm_expr.args[0] = new_lhs
        cpm_expr.update_args(cpm_expr.args) # XXX redundant? we know it's flat so no subexprs anyway
        return cpm_expr, new_cons
    
    return cpm_expr, []





"""
    Re-impementation of MUS-computation techniques in CPMPy

    - Deletion-based MUS
    - QuickXplain
    - Optimal MUS
"""
import warnings
import numpy as np
import cpmpy as cp
from cpmpy.transformations.get_variables import get_variables
from cpmpy.transformations.normalize import toplevel_list

from .utils import make_assump_model
from ...expressions.utils import is_num


def mus(soft, hard=[], solver="ortools"):
    """
        A CP deletion-based MUS algorithm using assumption variables
        and unsat core extraction

        For solvers that support s.solve(assumptions=...) and s.get_core()

        Each constraint is an arbitrary CPMpy expression, so it can
        also be sublists of constraints (e.g. constraint groups),
        contain aribtrary nested expressions, global constraints, etc.

        Will extract an unsat core and then shrink the core further
        by repeatedly ommitting one assumption variable.

        :param: soft: soft constraints, list of expressions
        :param: hard: hard constraints, optional, list of expressions
        :param: solver: name of a solver, must support assumptions (e.g, "ortools", "exact", "z3" or "pysat")
    """

    assert hasattr(cp.SolverLookup.get(solver), "get_core"), f"mus requires a solver that supports assumption variables, use mus_naive with {solver} instead"

    # make assumption (indicator) variables and soft-constrained model
    (m, soft, assump) = make_assump_model(soft, hard=hard)
    s = cp.SolverLookup.get(solver, m)

    # create dictionary from assump to soft
    dmap = dict(zip(assump, soft))

    # setting all assump vars to true should be UNSAT
    assert not s.solve(assumptions=assump), "MUS: model must be UNSAT"
    core = set(s.get_core())  # start from solver's UNSAT core

    # deletion-based MUS
    # order so that constraints with many variables are tried and removed first
    for c in sorted(core, key=lambda c : -len(get_variables(dmap[c]))):
        if c not in core:
            continue # already removed
        core.remove(c)
        if s.solve(assumptions=list(core)) is True:
            core.add(c)
        else: # UNSAT, use new solver core (clause set refinement)
            core = set(s.get_core())

    return [dmap[avar] for avar in core]


def quickxplain(soft, hard=[], solver="ortools"):
    """
        Find a preferred minimal unsatisfiable subset of constraints, based on the ordering of constraints.

        A total order is imposed on the constraints using the ordering of `soft`.
        Constraints with lower index are preferred over ones with higher index

        Assumption-based implementation for solvers that support s.solve(assumptions=...) and s.get_core()
        More naive version available as `quickxplain_naive` to use with other solvers.

        :param: soft: list of soft constraints to find a preferred minimal unsatisfiable subset of
        :param: hard: list of hard constraints, will be added to the model before solving
        :param: solver: name of a solver, must support assumptions (e.g, "ortools", "exact", "z3" or "pysat")

        CPMpy implementation of the QuickXplain algorithm by Junker:
            Junker, Ulrich. "Preferred explanations and relaxations for over-constrained problems." AAAI-2004. 2004.
            https://cdn.aaai.org/AAAI/2004/AAAI04-027.pdf
    """

    assert hasattr(cp.SolverLookup.get(solver), "get_core"), f"quickxplain requires a solver that supports assumption variables, use quickxplain_naive with {solver} instead"

    model, soft, assump = make_assump_model(soft, hard)
    s = cp.SolverLookup.get(solver, model)

    assert s.solve(assumptions=assump) is False, "The model should be UNSAT!"
    dmap = dict(zip(assump, soft))

    # the recursive call
    def do_recursion(soft, hard, delta):

        if len(delta) != 0 and s.solve(assumptions=hard) is False:
            # conflict is in hard constraints, no need to recurse
            return []

        if len(soft) == 1:
            # conflict is not in hard constraints, but only 1 soft constraint
            return list(soft)  # base case of recursion

        split = len(soft) // 2  # determine split point
        more_preferred, less_preferred = soft[:split], soft[split:]  # split constraints into two sets

        # treat more preferred part as hard and find extra constants from less preferred
        delta2 = do_recursion(less_preferred, hard + more_preferred, more_preferred)
        # find which preferred constraints exactly
        delta1 = do_recursion(more_preferred, hard + delta2, delta2)
        return delta1 + delta2

    # optimization: find max index of solver core
    solver_core = frozenset(s.get_core())
    max_idx = max(i for i, a in enumerate(assump) if a in solver_core)

    core = do_recursion(list(assump)[:max_idx + 1], [], [])
    return [dmap[a] for a in core]


def optimal_mus(soft, hard=[], weights=None, solver="ortools", hs_solver="ortools", do_solution_hint=True):
    """
        Find an optimal MUS according to a linear objective function.
        By not providing a weightvector, this function will return the smallest mus.
        Works by iteratively generating correction subsets and computing optimal hitting sets to those enumerated sets.
        For better performance of the algorithm, use an incemental solver to compute the hitting sets such as Gurobi.

        Assumption-based implementation for solvers that support s.solve(assumptions=...)
        More naive version available as `optimal_mus_naive` to use with other solvers.

        :param: soft: list of soft constraints to find an optimal MUS of
        :param: hard: list of hard constraints, will be added to the model before solving
        :param: solver: name of a solver, must support assumptions (e.g, "ortools", "exact", "z3" or "pysat")
        :param: hs_solver: the hitting-set solver to use, ideally incremental such as "gurobi"
        :param: do_solution_hint: when true, will favor large satisfiable subsets generated by the SAT-solver

        CPMpy implementation loosely based on the "OCUS" algorithm from:

            Gamba, Emilio, Bart Bogaerts, and Tias Guns. "Efficiently explaining CSPs with unsatisfiable subset optimization."
            Journal of Artificial Intelligence Research 78 (2023): 709-746.

    """
    assert hasattr(cp.SolverLookup.get(solver), "get_core"), f"optimal_mus requires a solver that supports assumption variables, use optimal_mus_naive with {solver} instead"
    
    model, soft, assump = make_assump_model(soft, hard)
    dmap = dict(zip(assump, soft)) # map assumption variables to constraints

    s = cp.SolverLookup.get(solver, model)
    if do_solution_hint and hasattr(s, 'solution_hint'): # algo is constructive, so favor large subsets
        s.solution_hint(assump, [1]*len(assump))

    assert s.solve(assumptions=assump) is False

    # initialize hitting set solver
    if weights is None:
        weights = np.ones(len(assump), dtype=int)

    hs_solver = cp.SolverLookup.get(hs_solver)
    hs_solver.minimize(cp.sum(assump * np.array(weights)))

    while hs_solver.solve() is True:
        hitting_set = [a for a in assump if a.value()]
        if s.solve(assumptions=hitting_set) is False:
            break

        # else, the hitting set is SAT, now try to extend it without extra solve calls.
        # Check which other assumptions/constraints are satisfied (using c.value())
        # complement of grown subset is a correction subset
        # Assumptions encode indicator constraints a -> c, find all false assumptions
        #   that really have to be false given the current solution.
        new_corr_subset = [a for a,c in zip(assump, soft) if a.value() is False and c.value() is False]
        hs_solver += cp.sum(new_corr_subset) >= 1

        # greedily search for other corr subsets disjoint to this one
        sat_subset = list(new_corr_subset)
        while s.solve(assumptions=sat_subset) is True:
            new_corr_subset = [a for a,c in zip(assump, soft) if a.value() is False and c.value() is False]
            sat_subset += new_corr_subset # extend sat subset with new corr subset, guaranteed to be disjoint
            hs_solver += cp.sum(new_corr_subset) >= 1 # add new corr subset to hitting set solver

    return [dmap[a] for a in hitting_set]

def smus(soft, hard=[], solver="ortools", hs_solver="ortools"):
    return optimal_mus(soft, hard=hard, weights=None, solver=solver, hs_solver=hs_solver)


## Naive, non-assumption based versions of MUS-algos above
def mus_naive(soft, hard=[], solver="ortools"):
    """
        A naive pure CP deletion-based MUS algorithm

        Will repeatedly solve the problem from scratch with one less constraint
        For anything but tiny sets of constraints, this will be terribly slow.

        Best to only use for testing on solvers that do not support assumptions.
        For others, use `mus()`

        :param soft: soft constraints, list of expressions
        :param hard: hard constraints, optional, list of expressions
        :param solver: name of a solver, see SolverLookup.solvernames()
    """
    # ensure toplevel list
    soft = toplevel_list(soft, merge_and=False)

    m = cp.Model(hard + soft)
    assert not m.solve(solver=solver), "MUS: model must be UNSAT"

    mus = []
    # order so that constraints with many variables are tried and removed first
    core = sorted(soft, key=lambda c: -len(get_variables(c)))
    for i in range(len(core)):
        subcore = mus + core[i + 1:]  # check if all but 'i' makes core SAT

        if cp.Model(hard + subcore).solve(solver=solver):
            # removing it makes it SAT, must keep for UNSAT
            mus.append(core[i])
        # else: still UNSAT so don't need this candidate

    return mus


def quickxplain_naive(soft, hard=[], solver="ortools"):
    """
        Find a preferred minimal unsatisfiable subset of constraints, based on the ordering of constraints.

        A total order is imposed on the constraints using the ordering of `soft`.
        Constraints with lower index are preferred over ones with higher index

        Naive implementation, re-solving the model from scratch.
        Can be slower depending on the number of global constraints used and solver support for reified constraints.

        CPMpy implementation of the QuickXplain algorithm by Junker:
            Junker, Ulrich. "Preferred explanations and relaxations for over-constrained problems." AAAI-2004. 2004.
            https://cdn.aaai.org/AAAI/2004/AAAI04-027.pdf
    """

    soft = toplevel_list(soft, merge_and=False)
    assert cp.Model(hard + soft).solve(solver) is False, "The model should be UNSAT!"

    # the recursive call
    def do_recursion(soft, hard, delta):

        m = cp.Model(hard)
        if len(delta) != 0 and m.solve(solver) is False:
            # conflict is in hard constraints, no need to recurse
            return []

        if len(soft) == 1:
            # conflict is not in hard constraints, but only 1 soft constraint
            return list(soft)  # base case of recursion

        split = len(soft) // 2  # determine split point
        more_preferred, less_preferred = soft[:split], soft[split:]  # split constraints into two sets

        # treat more preferred part as hard and find extra constants from less preferred
        delta2 = do_recursion(less_preferred, hard + more_preferred, more_preferred)
        # find which preferred constraints exactly
        delta1 = do_recursion(more_preferred, hard + delta2, delta2)
        return delta1 + delta2

    core = do_recursion(soft, hard, [])
    return core

def optimal_mus_naive(soft, hard=[], weights=None, solver="ortools", hs_solver="ortools"):
    """
        Naive implementation of `optimal_mus` without assumption variables and incremental solving
    """

    soft = toplevel_list(soft, merge_and=False)
    bvs = cp.boolvar(shape=len(soft))

    if weights is None:
        weights = np.ones(len(bvs), dtype=int)
    hs_solver = cp.SolverLookup.get(hs_solver)
    hs_solver.minimize(cp.sum(bvs * np.array(weights)))

    while hs_solver.solve() is True:

        hitting_set = [c for bv, c in zip(bvs, soft) if bv.value()]
        if cp.Model(hard + hitting_set).solve(solver=solver) is False:
            break

        # else, the hitting set is SAT, now try to extend it without extra solve calls.
        # Check which other assumptions/constraints are satisfied using its value() function
        #       sidenote: some vars may not be know to model and are None!
        # complement of grown subset is a correction subset
        false_constraints = [s for s in soft if s.value() is False or s.value() is None]
        corr_subset = [bv for bv,c in zip(bvs, soft) if c in frozenset(false_constraints)]
        hs_solver += cp.sum(corr_subset) >= 1

        # find more corr subsets, disjoint to this one
        sat_subset = hitting_set + false_constraints
        while cp.Model(hard + sat_subset).solve(solver=solver) is True:
            false_constraints = [s for s in soft if s.value() is False or s.value() is None]
            corr_subset = [bv for bv, c in zip(bvs, soft) if c in frozenset(false_constraints)]
            hs_solver += cp.sum(corr_subset) >= 1
            sat_subset += false_constraints

    return hitting_set







#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## hexaly.py
##
"""
    Interface to Hexaly's API


    Hexaly is a local search solver with support for  global constraints.

    Always use :func:`cp.SolverLookup.get("hexaly") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'hexaly' python package is installed:

    .. code-block:: console

        $ pip install hexaly -i https://pip.hexaly.com                
    
    The Hexaly local solver requires an active licence (for example a free academic license)
    You can read more about available licences at https://www.hexaly.com/

    See detailed installation instructions at:
    https://www.hexaly.com/docs/last/installation/pythonsetup.html

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_hexaly
"""

from typing import Optional

from importlib.metadata import version, PackageNotFoundError

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions.core import Expression, Comparison, Operator, BoolVal
from ..expressions.globalconstraints import GlobalConstraint, GlobalFunction, DirectConstraint
from ..expressions.variables import _BoolVarImpl, NegBoolView, _IntVarImpl, _NumVarImpl
from ..expressions.utils import is_num, is_any_list, eval_comparison, flatlist
from ..transformations.get_variables import get_variables
from ..transformations.normalize import toplevel_list
from ..transformations.decompose_global import decompose_in_tree

class CPM_hexaly(SolverInterface):
    """
    Interface to Hexaly's API

    Creates the following attributes (see parent constructor for more):

    - hex_model: object, Hexaly's model object
    - hex_solver: object, Hexaly's solver object (to solve hex_model)

    Documentation of the solver's own Python API:
    https://www.hexaly.com/docs/last/pythonapi/index.html
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            import hexaly as hex
            return True
        except ModuleNotFoundError: # if solver's Python package is not installed
            return False
        except Exception as e:
            raise e

    @classmethod
    def version(cls) -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return version('hexaly')
        except PackageNotFoundError:
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
        - cpm_model: Model(), a CPMpy Model() (optional)
        - subsolver: str, name of a subsolver (optional)
        """
        if not self.supported():
            raise Exception("CPM_hexaly: Install the python package 'hexaly' to use this solver interface.")

        from hexaly.optimizer import HexalyOptimizer

        assert subsolver is None # hexaly does not have subsolvers

        # initialise the native solver object
        self.hex_solver = HexalyOptimizer()
        self.hex_solver.param.verbosity = 0
        self.hex_model = self.hex_solver.model
        self.is_satisfaction = True

        # initialise everything else and post the constraints/objective
        super().__init__(name="hexaly", cpm_model=cpm_model)

    @property
    def native_model(self):
        return self.hex_model

    def solve(self, time_limit=None, **kwargs):
        """
            Call the Hexaly solver

            Arguments:
                time_limit:  maximum solve time in seconds (float, optional)
                kwargs:      any keyword argument, sets parameters of solver object

            Arguments that correspond to solver parameters:

            - nb_threads: number of threads used to parallelize the search.
            - iteration_limit: max number of iterations
            - verbosity: verbosity level

            full list of parameters availble at:
            https://www.hexaly.com/docs/last/pythonapi/optimizer/hxparam.html
        """
        from hexaly.optimizer import HxObjectiveDirection

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        if time_limit is not None:
            if time_limit < 0:
                raise ValueError(f"Time limit must be positive but was {time_limit}")
            self.hex_solver.param.time_limit = int(time_limit)  # hexaly does not support float time limit

        # set solver parameters
        for arg, val in kwargs.items():
            setattr(self.hex_solver, arg, val)

        if self.is_satisfaction: # set dummy objective for satisfaction problems
            self.hex_model.add_objective(0, HxObjectiveDirection.MINIMIZE)

        # new status, translate runtime
        self.hex_model.close() # model must be closed
        self.hex_solver.solve()
        self.hex_sol = self.hex_solver.get_solution()
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.hex_solver.statistics.running_time # wallclock time in (float) seconds

        # Translate solver exit status to CPMpy exit status
        # CSP:                         COP:
        # ├─ sat -> FEASIBLE           ├─ optimal -> OPTIMAL
        # ├─ unsat -> UNSATISFIABLE    ├─ sub-optimal -> FEASIBLE
        # └─ timeout -> UNKNOWN        ├─ unsat -> UNSATISFIABLE
        #                              └─ timeout -> UNKNOWN

        from hexaly.optimizer import HxSolutionStatus
        if self.hex_sol.status == HxSolutionStatus.INCONSISTENT:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif self.hex_sol.status == HxSolutionStatus.INFEASIBLE:
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        elif self.hex_sol.status == HxSolutionStatus.FEASIBLE:
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif self.hex_sol.status == HxSolutionStatus.OPTIMAL:
            if self.is_satisfaction:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            else:
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
        else:  # another?
            raise NotImplementedError(self.hex_sol.status)  # a new status type was introduced, please report on github

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:
            # fill in variable values
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if cpm_var.is_bool():
                    cpm_var._value = bool(self.hex_sol.get_value(sol_var))
                else:
                    cpm_var._value = int(self.hex_sol.get_value(sol_var))

            # translate objective, for optimisation problems only
            if not self.is_satisfaction:
                self.objective_value_ = self.hex_sol.get_objective_bound(0)

        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var.clear()

        # now open model again, we might want to add new constraints after
        self.hex_model.open()

        if self.is_satisfaction:
            self.hex_model.remove_objective(0) # reset to not have any objectives

        return has_sol


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        if is_num(cpm_var): # shortcut, eases posting constraints
            return cpm_var

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            return ~self.solver_var(cpm_var._bv)

        # create if it does not exist
        if cpm_var not in self._varmap:
            if isinstance(cpm_var, _BoolVarImpl):
                revar = self.hex_model.bool()
            elif isinstance(cpm_var, _IntVarImpl):
                revar = self.hex_model.int(cpm_var.lb, cpm_var.ub)
            else:
                raise NotImplementedError("Not a known var {}".format(cpm_var))
            # set name of variable
            revar.set_name(str(cpm_var))
            self._varmap[cpm_var] = revar

        # return from cache
        return self._varmap[cpm_var]


    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize

            'objective()' can be called multiple times, only the last one is stored

            (technical side note: any constraints created during conversion of the objective

            are permanently posted to the solver)
        """
        from hexaly.optimizer import HxObjectiveDirection
        # make objective function or variable and post
        while self.has_objective(): # remove prev objective(s)
            self.hex_model.remove_objective(0)
        self.is_satisfaction = False
        hex_obj = self._hex_expr(expr)
        if minimize:
            self.hex_model.add_objective(hex_obj,HxObjectiveDirection.MINIMIZE)
        else:
            self.hex_model.add_objective(hex_obj,HxObjectiveDirection.MAXIMIZE)

    def has_objective(self):
        return self.hex_model.nb_objectives > 0

    # `add()` first calls `transform()`
    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: list of Expression
        """
        # apply transformations
        cpm_cons = toplevel_list(cpm_expr)
        # no flattening, so also no safening required
        cpm_cons = decompose_in_tree(cpm_cons, supported={"min", "max", "abs", "element"})
        return cpm_cons

    def add(self, cpm_expr_orig):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: self
        """
        # add new user vars to the set
        get_variables(cpm_expr_orig, collect=self.user_vars)

        # transform and post the constraints
        for cpm_expr in self.transform(cpm_expr_orig):
            hex_expr = self._hex_expr(cpm_expr)
            self.hex_model.add_constraint(hex_expr)

        return self
    __add__ = add  # avoid redirect in superclass

    def _hex_expr(self, cpm_expr):

        # get transformed constraint

        if is_any_list(cpm_expr):
            return [self._hex_expr(expr) for expr in cpm_expr]

        # constants
        if isinstance(cpm_expr, BoolVal):
            return bool(cpm_expr)
        if is_num(cpm_expr):
            return cpm_expr
        # variables
        if isinstance(cpm_expr, _NumVarImpl):
            return self.solver_var(cpm_expr)

        if isinstance(cpm_expr, Operator):
            if cpm_expr.name == "and":
                return self.hex_model.and_(self._hex_expr(cpm_expr.args))
            if cpm_expr.name == "or":
                return self.hex_model.or_(self._hex_expr(cpm_expr.args))
            if cpm_expr.name == "not":
                return ~self._hex_expr(cpm_expr.args[0])
            if cpm_expr.name == "->":
                cond, subexpr = cpm_expr.args
                return self._hex_expr(~cond | subexpr) # post as disjunction
            if cpm_expr.name == "sum":
                return self.hex_model.sum(self._hex_expr(cpm_expr.args))
            if cpm_expr.name == "wsum":
                weights, args = cpm_expr.args
                return self.hex_model.sum([w * a for w,a in zip(weights, self._hex_expr(args))])
            if cpm_expr.name == "sub":
                a,b = self._hex_expr(cpm_expr.args)
                return a - b
            if cpm_expr.name == "-":
                return -self._hex_expr(cpm_expr.args[0])
            if cpm_expr.name == "mul":
                a,b = self._hex_expr(cpm_expr.args)
                return a * b
            if cpm_expr.name == "div":
                a, b = self._hex_expr(cpm_expr.args)
                # ensure we are rounding towards zero
                return self.hex_model.iif((a >= 0) & (b >= 0), self.hex_model.floor(a / b), # result is positive
                       self.hex_model.iif((a <= 0) & (b <= 0), self.hex_model.floor(a / b), # result is positive
                       self.hex_model.iif((a >= 0) & (b <= 0), self.hex_model.ceil(a / b), # result is negative
                       self.hex_model.iif((a <= 0) & (b >= 0), self.hex_model.ceil(a / b), 0)))) # result is negative

            if cpm_expr.name == "mod":
                a, b = self._hex_expr(cpm_expr.args)
                return a % b
            if cpm_expr.name == "pow":
                a, b = self._hex_expr(cpm_expr.args)
                return a ** b
            raise ValueError(f"Unknown operator {cpm_expr}")

        elif isinstance(cpm_expr, Comparison):
            x,y = self._hex_expr(cpm_expr.args)
            return eval_comparison(cpm_expr.name, x,y)

        elif isinstance(cpm_expr, GlobalConstraint):
            if cpm_expr.name == "alldifferent":
                hex_arr = self.hex_model.array(self._hex_expr(cpm_expr.args))
                return self.hex_model.distinct(hex_arr)
            raise ValueError(f"Global constraint {cpm_expr} is not supported by hexaly")

        elif isinstance(cpm_expr, GlobalFunction):
            if cpm_expr.name == "nvalues":
                return self.hex_model.distinct(self._hex_expr(cpm_expr.args))
            if cpm_expr.name == "element":
                hex_arr = self.hex_model.array(self._hex_expr(cpm_expr.args[0]))
                idx = self._hex_expr(cpm_expr.args[1])
                return self.hex_model.at(hex_arr,idx)
            if cpm_expr.name == "abs":
                return self.hex_model.abs(self._hex_expr(cpm_expr.args[0]))
            if cpm_expr.name == "min":
                return self.hex_model.min(*self._hex_expr(cpm_expr.args))
            if cpm_expr.name == "max":
                return self.hex_model.max(*self._hex_expr(cpm_expr.args))
            raise ValueError(f"Global function {cpm_expr} is not supported by hexaly")

        elif isinstance(cpm_expr, DirectConstraint):
            return cpm_expr.callSolver(self, self.hex_model)

        raise NotImplementedError(f"Unexpected expression {cpm_expr}")

    
    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            A shorthand to (efficiently) compute all solutions, map them to CPMpy and optionally display the solutions.

            Arguments:
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                        default/None: nothing displayed
                solution_limit: stop after this many solutions (default: None)
                time_limit (float):   maximum solve time in seconds

            Returns: 
                number of solutions found

            .. note::
                Hexaly does not support exhaustive search to find all solutions. Set `time_limit` to do a limited search.
        """
        if time_limit is None:
            raise ValueError("Hexaly does not support exhaustive search to find all solutions. "
                             "Set time limit to do a limited search")

        return super(CPM_hexaly, self).solveAll(display, time_limit, solution_limit, call_from_model, **kwargs)


    def solution_hint(self, cpm_vars, vals):
        from hexaly.optimizer import HxObjectiveDirection
        if self.is_satisfaction: # set dummy objective, otherwise cannot close model
            self.hex_model.add_objective(0, HxObjectiveDirection.MINIMIZE)

        cpm_vars = flatlist(cpm_vars)
        vals = flatlist(vals)

        self.hex_model.close() # must be closed before we can set a solution hint
        for hex_var, val in zip(self.solver_vars(cpm_vars), vals):
            hex_var.value = val
        self.hex_model.open() # re-open

    def __del__(self):
        # release lock on licence file
        self.hex_solver.delete()




#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## dimacs.py
##
"""
    This file implements helper functions for exporting CPMpy models from and to DIMACS format.
    DIMACS is a textual format to represent CNF problems.
    The header of the file should be formatted as ``p cnf <n_vars> <n_constraints>``.
    If the number of variables and constraints are not given, it is inferred by the parser.

    Each remaining line of the file is formatted as a list of integers.
    An integer represents a Boolean variable and a negative Boolean variable is represented using a `'-'` sign.
"""

import cpmpy as cp

from cpmpy.expressions.variables import _BoolVarImpl, NegBoolView
from cpmpy.expressions.core import Operator, Comparison

from cpmpy.transformations.normalize import toplevel_list
from cpmpy.transformations.to_cnf import to_cnf
from cpmpy.transformations.get_variables import get_variables

import re


def write_dimacs(model, fname=None):
    """
        Writes CPMpy model to DIMACS format
        Uses the "to_cnf" transformation from CPMpy

        .. todo::
            TODO: implement pseudoboolean constraints in to_cnf

        :param model: a CPMpy model
        :param fname: optional, file name to write the DIMACS output to
    """

    constraints = toplevel_list(model.constraints)
    constraints = to_cnf(constraints)

    vars = get_variables(constraints)
    mapping = {v : i+1 for i, v in enumerate(vars)}

    out = f"p cnf {len(vars)} {len(constraints)}\n"
    for cons in constraints:

        if isinstance(cons, _BoolVarImpl):
            cons = Operator("or", [cons])

        if not (isinstance(cons, Operator) and cons.name == "or"):
            raise NotImplementedError(f"Unsupported constraint {cons}")

        # write clause to cnf format
        ints = []
        for v in cons.args:
            if isinstance(v, NegBoolView):
                ints.append(str(-mapping[v._bv]))
            elif isinstance(v, _BoolVarImpl):
                ints.append(str(mapping[v]))
            else:
                raise ValueError(f"Expected Boolean variable in clause, but got {v} which is of type {type(v)}")

        out += " ".join(ints + ["0"]) + "\n"

    if fname is not None:
        with open(fname, "w") as f:
            f.write(out)

    return out


def read_dimacs(fname):
    """
        Read a CPMpy model from a DIMACS formatted file strictly following the specification:
        https://web.archive.org/web/20190325181937/https://www.satcompetition.org/2009/format-benchmarks2009.html
        
        .. note::
            The p-line has to denote the correct number of variables and clauses
        
        :param fname: the name of the DIMACS file
        :param sep: optional, separator used in the DIMACS file, will try to infer if None
    """

    m = cp.Model()

    with open(fname, "r") as f:
        clause = []
        nr_vars = None
        for line in f.readlines():
            if line == "" or line.startswith("c"):
                continue  # skip empty and comment lines
            elif line.startswith("p"):
                params = line.strip().split(" ")
                assert len(params) == 4, f"Expected p-header to be formed `p cnf nr_vars nr_cls` but got {line}"
                _,typ,nr_vars,nr_cls = params
                if typ != "cnf":
                    raise ValueError("Expected `cnf` (i.e. DIMACS) as file format, but got {typ} which is not supported.")
                nr_vars = int(nr_vars)
                if nr_vars>0:
                    bvs = cp.boolvar(shape=nr_vars)
                nr_cls = int(nr_cls)
            else:
                assert nr_vars is not None, "Expected p-line before first clause"
                for token in line.strip().split():
                    i = int(token.strip())
                    if i == 0:
                        m += cp.any(clause)
                        clause = []
                    else:
                        var=abs(i)-1
                        assert var < nr_vars, "Expected at most {nr_vars} variables (from p-line) but found literal {i} in clause {line}"
                        bv = bvs[var]

                        clause.append(bv if i > 0 else ~bv)

        assert nr_vars is not None, "Expected file to contain p-line, but did not"
        assert len(clause) == 0, f"Expected last clause to be terminated by 0, but it was not"
        assert len(m.constraints) == nr_cls, f"Number of clauses was declared in p-line as {nr_cls}, but was {len(m.constraints)}"

    return m








#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## __init__.py
##
"""
    Set of utilities for working with XCSP3-formatted CP models.


    =================
    List of functions
    =================

    .. autosummary::
        :nosignatures:

        read_xcsp3

    ========================
    List of helper functions
    ========================

    .. autosummary::
        :nosignatures:

        _parse_xcsp3
        _load_xcsp3

    ==================
    List of submodules
    ==================

    .. autosummary::
        :nosignatures:

        parser_callbacks
        analyze
        benchmark
        xcsp3_cpmpy
        dataset
        globals
"""
from io import StringIO
import lzma
import os
import cpmpy as cp

# Special case for optional cpmpy dependencies
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pycsp3.parser.xparser import CallbackerXCSP3, ParserXCSP3

from .dataset import XCSP3Dataset # for easier importing

def _parse_xcsp3(path: os.PathLike) -> "ParserXCSP3":
    """
    Parses an XCSP3 instance file (.xml) and returns a `ParserXCSP3` instance.
    
    Arguments:
        path: location of the XCSP3 instance to read (expects a .xml file).
    
    Returns:
        A parser object.
    """
    try:
        from pycsp3.parser.xparser import ParserXCSP3
    except ImportError as e:
        raise ImportError("The 'pycsp3' package is required to parse XCSP3 files. "
                          "Please install it with `pip install pycsp3`.") from e
    
    parser = ParserXCSP3(path)
    return parser

def _load_xcsp3(parser: "ParserXCSP3") -> cp.Model:
    """
    Takes in a `ParserXCSP3` instance and loads its captured model as a CPMpy model.

    Arguments:
        parser (ParserXCSP3): A parser object to load from.

    Returns:
        The XCSP3 instance loaded as a CPMpy model.
    """
    from .parser_callbacks import CallbacksCPMPy
    from pycsp3.parser.xparser import CallbackerXCSP3
    callbacks = CallbacksCPMPy()
    callbacks.force_exit = True
    callbacker = CallbackerXCSP3(parser, callbacks)
    callbacker.load_instance()
    model = callbacks.cpm_model
   
    return model


def read_xcsp3(path: os.PathLike) -> cp.Model:
    """
    Reads in an XCSP3 instance (.xml or .xml.lzma) and returns its matching CPMpy model.

    Arguments:
        path: location of the XCSP3 instance to read (expects a .xml or .xml.lzma file).

    Returns:
        The XCSP3 instance loaded as a CPMpy model.
    """
    # Decompress on the fly if still in .lzma format
    if str(path).endswith(".lzma"):
        path = decompress_lzma(path)

    # Parse and create CPMpy model
    parser = _parse_xcsp3(path)
    model = _load_xcsp3(parser)
    return model

def decompress_lzma(path: os.PathLike) -> StringIO:
    """
    Decompresses a .lzma file.

    Arguments:
        path: Location of .lzma file

    Returns:
        Memory-mapped decompressed file
    """
    # Decompress the XZ file
    with lzma.open(path, 'rt', encoding='utf-8') as f:
        return StringIO(f.read()) # read to memory-mapped file
        

    


#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## pysdd.py
##
"""
    Interface to PySDD's API

    PySDD is a knowledge compilation package for Sentential Decision Diagrams (SDD).
    (see https://pysdd.readthedocs.io/en/latest/)

    .. warning::    
        This solver can ONLY be used for solution checking and enumeration over Boolean variables!
        It does not support optimization.

    Always use :func:`cp.SolverLookup.get("pysdd") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'PySDD' python package is installed:

    .. code-block:: console

        $ pip install PySDD

    See detailed installation instructions at:
    https://pysdd.readthedocs.io/en/latest/usage/installation.html

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_pysdd

    ==============
    Module details
    ==============
"""
from functools import reduce
from typing import Optional
import pkg_resources

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..exceptions import NotSupportedError
from ..expressions.core import Expression, BoolVal
from ..expressions.variables import _BoolVarImpl, NegBoolView, boolvar
from ..expressions.globalconstraints import DirectConstraint
from ..expressions.utils import is_bool, argval, argvals
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.get_variables import get_variables
from ..transformations.normalize import toplevel_list, simplify_boolean

class CPM_pysdd(SolverInterface):
    """
    Interface to PySDD's API.

    Creates the following attributes (see parent constructor for more):

    - ``pysdd_vtree`` : a pysdd.sdd.Vtree
    - ``pysdd_manager`` : a pysdd.sdd.SddManager
    - ``pysdd_root`` : a pysdd.sdd.SddNode (changes whenever a formula is added)

    The :class:`~cpmpy.expressions.globalconstraints.DirectConstraint`, when used, calls a function on the ``pysdd_manager`` object and replaces the root node with a conjunction of the previous root node and the result of this function call.

    Documentation of the solver's own Python API:
    https://pysdd.readthedocs.io/en/latest/classes/SddManager.html
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            from pysdd.sdd import SddManager
            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e
        
    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('pysdd').version
        except pkg_resources.DistributionNotFound:
            return None


    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Requires a CPMpy model as input, and will create the corresponding
        pysdd vtree, manager and (True) root node

        Only supports satisfaction problems and solution enumeration

        Arguments:
            cpm_model: Model(), a CPMpy Model(), optional
            subsolver: None
        """
        if not self.supported():
            raise Exception("CPM_pysdd: Install the python package 'pysdd' to use this solver interface")
        if cpm_model and cpm_model.objective_ is not None:
            raise NotSupportedError("CPM_pysdd: only satisfaction, does not support an objective function")

        from pysdd.sdd import SddManager, Vtree

        cnt = 1
        self.pysdd_vtree = Vtree(var_count=cnt, vtree_type="balanced")
        self.pysdd_manager = SddManager.from_vtree(self.pysdd_vtree)
        self.pysdd_root = self.pysdd_manager.true()

        # initialise everything else and post the constraints/objective
        super().__init__(name="pysdd", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.pysdd_root

    def solve(self, time_limit=None, assumptions=None):
        """
            See if an arbitrary model exists

            This is a knowledge compiler:

            - building it is the (computationally) hard part
            - checking for a solution is trivial after that
        """

        if time_limit is not None:
            raise NotImplementedError("PySDD.solve(), time_limit not (yet?) supported")
        
        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(boolvar() == True)

        has_sol = True
        if self.pysdd_root is not None:
            # if root node is false (empty), no solutions
            has_sol = not self.pysdd_root.is_false()

        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = 0.0

        # translate exit status
        if has_sol:
            # Only CSP (does not support COP)
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        else:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
            for cpm_var in self.user_vars:
                cpm_var._value = None

        # get solution values (of user specified variables only)
        if has_sol and self.pysdd_root is not None:
            sol = next(self.pysdd_root.models())
            # fill in variable values
            for cpm_var in self.user_vars:
                lit = self.solver_var(cpm_var).literal
                if lit in sol:
                    cpm_var._value = bool(sol[lit])
                else:
                    cpm_var._value = cpm_var.get_bounds()[0] # dummy value - TODO: ensure Pysdd assigns an actual value
                    # cpm_var._value = None  # not specified...

        return has_sol

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            Compute all solutions and optionally display the solutions.

            .. warning::
                WARNING: setting 'display' will SIGNIFICANTLY slow down solution counting...

            Arguments:
                - display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                        default/None: nothing displayed
                - time_limit, solution_limit, kwargs: not used
                - call_from_model: whether the method is called from a CPMpy Model instance or not

            Returns: 
                number of solutions found            
        """
        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(boolvar() == True)

        if time_limit is not None:
            raise NotImplementedError("PySDD.solveAll(), time_limit not (yet?) supported")
        if solution_limit is not None:
            raise NotImplementedError("PySDD.solveAll(), solution_limit not (yet?) supported")

        if self.pysdd_root is None or self.pysdd_root.is_false():
            # clear user vars if no solution found
            for var in self.user_vars:
                var._value = None
            return 0

        sddmodels = [x for x in self.pysdd_root.models()]
        if len(sddmodels) != self.pysdd_root.model_count:
            #pysdd doesn't always have correct solution count..
            projected_sols = set()
            for sol in sddmodels:
                projectedsol = []
                for cpm_var in self.user_vars:
                    lit = self.solver_var(cpm_var).literal
                    projectedsol.append(bool(sol[lit]))
                projected_sols.add(tuple(projectedsol))
        else:
            projected_sols = set(sddmodels)

        if projected_sols:
            if projected_sols == solution_limit:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            else:
                # time limit not (yet) supported -> always all solutions found
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
        else:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE

        # display if needed
        if display is not None:
            # manually walking over the tree, much slower...
            for sol in projected_sols:
                # fill in variable values
                for i, cpm_var in enumerate(self.user_vars):
                    cpm_var._value = sol[i]

                if isinstance(display, Expression):
                    print(argval(display))
                elif isinstance(display, list):
                    print(argvals(display))
                else:
                    display()  # callback
        
        return len(projected_sols)

    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
        """
        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            # just a view, get actual var identifier, return -id
            return -self.solver_var(cpm_var._bv)

        # create if it does not exist
        if cpm_var not in self._varmap:
            if isinstance(cpm_var, _BoolVarImpl):
                # make new var, add at end (what is best here??)
                self.pysdd_manager.add_var_after_last()
                n = self.pysdd_manager.var_count()
                revar = self.pysdd_manager.vars[n]
            else:
                raise NotImplementedError(f"CPM_pysdd: non-Boolean variable {cpm_var} not supported")
            self._varmap[cpm_var] = revar

        return self._varmap[cpm_var]

    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the ':ref:`Adding a new solver` docs on readthedocs for more information.

            For PySDD, it can be beneficial to add a big model (collection of constraints) at once...

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        # works on list of nested expressions
        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = decompose_in_tree(cpm_cons,supported={'xor'}, supported_reified={'xor'}, csemap=self._csemap) #keep unsupported xor for error message purposes.
        cpm_cons = simplify_boolean(cpm_cons)  # for cleaning (BE >= 0) and such
        return cpm_cons

    def add(self, cpm_expr):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: self
        """

        newvars = get_variables(cpm_expr)

        # add new user vars to the set
        self.user_vars |= set(newvars)

        # transform and post the constraints
        # XXX the order in the for loop will matter on runtime efficiency...
        for cpm_con in self.transform(cpm_expr):
            # replace root by conjunction of itself and the con expression
            self.pysdd_root = self.pysdd_manager.conjoin(self.pysdd_root,
                                                self._pysdd_expr(cpm_con))

        return self
    __add__ = add  # avoid redirect in superclass

    def _pysdd_expr(self, cpm_con):
        """
            PySDD supports nested expressions: each expression
            (variable or subexpression) is a node...
            so we recursively translate our expressions to theirs.

            input: Expression or const
            output: pysdd Node
        """
        if isinstance(cpm_con, _BoolVarImpl):
            # base case, just var or ~var
            return self.solver_var(cpm_con)

        elif is_bool(cpm_con) or isinstance(cpm_con, BoolVal):
            # base case: Boolean value
            if cpm_con:
                return self.pysdd_manager.true()
            else:
                return self.pysdd_manager.false()

        elif not isinstance(cpm_con, Expression):
            # a number or so
            raise NotImplementedError(f"CPM_pysdd: Non supported object {cpm_con}")

        elif cpm_con.name == 'and':
            # conjoin the nodes corresponding to the args
            # also here the order might matter on runtime efficiency...
            return reduce(self.pysdd_manager.conjoin, [self._pysdd_expr(a) for a in cpm_con.args])

        elif cpm_con.name == 'or':
            # disjoin the nodes corresponding to the args
            # also here the order might matter on runtime efficiency...
            return reduce(self.pysdd_manager.disjoin, [self._pysdd_expr(a) for a in cpm_con.args])

        elif cpm_con.name == 'not':
            return self.pysdd_manager.negate(self._pysdd_expr(cpm_con.args[0]))

        elif cpm_con.name == '->':
            a0 = self._pysdd_expr(cpm_con.args[0])
            a1 = self._pysdd_expr(cpm_con.args[1])
            # ~a0 | a1
            return self.pysdd_manager.disjoin(self.pysdd_manager.negate(a0), a1)

        elif cpm_con.name == '==':
            a0 = self._pysdd_expr(cpm_con.args[0])
            a1 = self._pysdd_expr(cpm_con.args[1])
            # (~a0 | a1) & (~a1 | a0)
            return self.pysdd_manager.conjoin(
                        self.pysdd_manager.disjoin(self.pysdd_manager.negate(a0), a1),
                        self.pysdd_manager.disjoin(self.pysdd_manager.negate(a1), a0),
                   )

        elif cpm_con.name == '!=':
            # ~(a0 == a1)
            equiv = self._pysdd_expr(cpm_con.args[0] == cpm_con.args[1])
            return self.pysdd_manager.negate(equiv)

        # a direct constraint, call on manager
        # WARNING: will only work when all args are variables or constants!
        # if unwanted, repeated some of the logic of callSolver here
        elif isinstance(cpm_con, DirectConstraint):
            return cpm_con.callSolver(self, self.pysdd_manager)

        else:
            raise NotImplementedError(f"CPM_pysdd: Non supported constraint {cpm_con}")

    def dot(self):
        """
            Returns a graphviz Dot object

            Display (in a notebook) with:

            .. code-block:: python

                import graphviz
                graphviz.Source(m.dot())
        """
        if self.pysdd_root is None:
            from pysdd.sdd import SddManager
            SddManager().true().dot()
        return self.pysdd_root.dot()



"""
    This file implements parameter tuning for constraint solvers based on SMBO and using adaptive capping.
    Based on the following paper:
    Ignace Bleukx, Senne Berden, Lize Coenen, Nicholas Decleyre, Tias Guns (2022). Model-Based Algorithm
    Configuration with Adaptive Capping and Prior Distributions. In: Schaus, P. (eds) Integration of Constraint
    Programming, Artificial Intelligence, and Operations Research. CPAIOR 2022. Lecture Notes in Computer Science,
    vol 13292. Springer, Cham. https://doi.org/10.1007/978-3-031-08011-1_6

    - DOI: https://doi.org/10.1007/978-3-031-08011-1_6
    - Link to paper: https://rdcu.be/cQyWR
    - Link to original code of paper: https://github.com/ML-KULeuven/DeCaprio

    This code currently only implements the author's 'Hamming' surrogate function.
    The parameter tuner iteratively finds better hyperparameters close to the current best configuration during the search.
    Searching and time-out start at the default configuration for a solver (if available in the solver class)
"""
import time
from random import shuffle

import numpy as np

from ..solvers.utils import SolverLookup, param_combinations
from ..solvers.solver_interface import ExitStatus

class ParameterTuner:
    """
        Parameter tuner based on DeCaprio method [ref_to_decaprio]
    """

    def __init__(self, solvername, model, all_params=None, defaults=None):
        """            
            :param solvername: Name of solver to tune
            :param model: CPMpy model to tune parameters on
            :param all_params: optional, dictionary with parameter names and values to tune. If None, use predefined parameter set.
        """
        self.solvername = solvername
        self.model = model
        self.all_params = all_params
        self.best_params = defaults
        if self.all_params is None:
            self.all_params = SolverLookup.lookup(solvername).tunable_params()
            self.best_params = SolverLookup.lookup(solvername).default_params()

        self._param_order = list(self.all_params.keys())
        self._best_config = self._params_to_np([self.best_params])

    def tune(self, time_limit=None, max_tries=None, fix_params={}):
        """
            :param time_limit: Time budget to run tuner in seconds. Solver will be interrupted when time budget is exceeded
            :param max_tries: Maximum number of configurations to test
            :param fix_params: Non-default parameters to run solvers with.
        """
        if time_limit is not None:
            start_time = time.time()

        # Init solver
        solver = SolverLookup.get(self.solvername, self.model)
        solver.solve(**self.best_params)

        self.base_runtime = solver.status().runtime
        self.best_runtime = self.base_runtime

        # Get all possible hyperparameter configurations
        combos = list(param_combinations(self.all_params))
        combos_np = self._params_to_np(combos)

        # Ensure random start
        np.random.shuffle(combos_np)

        i = 0
        if max_tries is None:
            max_tries = len(combos_np)
        while len(combos_np) and i < max_tries:
            # Make new solver
            solver = SolverLookup.get(self.solvername, self.model)
            # Apply scoring to all combos
            scores = self._get_score(combos_np)
            max_idx = np.where(scores == scores.min())[0][0]
            # Get index of optimal combo
            params_np = combos_np[max_idx]
            # Remove optimal combo from combos
            combos_np = np.delete(combos_np, max_idx, axis=0)
            # Convert numpy array back to dictionary
            params_dict = self._np_to_params(params_np)
            # set fixed params
            params_dict.update(fix_params)
            timeout = self.best_runtime
            # set timeout depending on time budget
            if time_limit is not None:
                timeout = min(timeout, time_limit - (time.time() - start_time))
            # run solver
            solver.solve(**params_dict, time_limit=timeout)
            if solver.status().exitstatus == ExitStatus.OPTIMAL and  solver.status().runtime < self.best_runtime:
                self.best_runtime = solver.status().runtime
                # update surrogate
                self._best_config = params_np

            if time_limit is not None and (time.time() - start_time) >= time_limit:
                break
            i += 1

        self.best_params = self._np_to_params(self._best_config)
        self.best_params.update(fix_params)
        return self.best_params

    def _get_score(self, combos):
        """
            Return the hamming distance for each remaining configuration to the current best config.
            Lower score means better configuration, so exploit the current best configuration by only allowing small changes.
        """
        return np.count_nonzero(combos != self._best_config, axis=1)

    def _params_to_np(self,combos):
        arr = [[params[key] for key in self._param_order] for params in combos]
        return np.array(arr)

    def _np_to_params(self,arr):
        return {key: val for key, val in zip(self._param_order, arr)}



class GridSearchTuner(ParameterTuner):

    def __init__(self, solvername, model, all_params=None, defaults=None):
        super().__init__(solvername, model, all_params, defaults)

    def tune(self, time_limit=None, max_tries=None, fix_params={}):
        """
            :param time_limit: Time budget to run tuner in seconds. Solver will be interrupted when time budget is exceeded
            :param max_tries: Maximum number of configurations to test
            :param fix_params: Non-default parameters to run solvers with.
        """
        if time_limit is not None:
            start_time = time.time()

        # Init solver
        solver = SolverLookup.get(self.solvername, self.model)
        solver.solve(**self.best_params)


        self.base_runtime = solver.status().runtime
        self.best_runtime = self.base_runtime

        # Get all possible hyperparameter configurations
        combos = list(param_combinations(self.all_params))
        shuffle(combos) # test in random order

        if max_tries is not None:
            combos = combos[:max_tries]

        for params_dict in combos:
            # Make new solver
            solver = SolverLookup.get(self.solvername, self.model)
            # set fixed params
            params_dict.update(fix_params)
            timeout = self.best_runtime
            # set timeout depending on time budget
            if time_limit is not None:
                timeout = min(timeout, time_limit - (time.time() - start_time))
            # run solver
            solver.solve(**params_dict, time_limit=timeout)
            if solver.status().exitstatus == ExitStatus.OPTIMAL and solver.status().runtime < self.best_runtime:
                self.best_runtime = solver.status().runtime
                # update surrogate
                self.best_params = params_dict

            if time_limit is not None and (time.time() - start_time) >= time_limit:
                break

        return self.best_params






#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## pumpkin.py
##
"""
    Interface to Pumpkin's API

    Pumpkin is a combinatorial optimisation solver developed by the ConSol Lab at TU Delft. 
    It is based on the (lazy clause generation) constraint programming paradigm.
    (see https://github.com/consol-lab/pumpkin)

    Always use :func:`cp.SolverLookup.get("pumpkin") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ===============
    Installation
    ===============

    Requires that the 'pumpkin-solver' python package is installed:

    .. code-block:: console
    
        $ pip install pumpkin-solver

    The rest of this documentation is for advanced users

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_pumpkin

    ==============
    Module details
    ==============
"""
from typing import Optional
from importlib.metadata import version, PackageNotFoundError
from os.path import join

import numpy as np

from cpmpy.exceptions import NotSupportedError
from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions.core import Expression, Comparison, Operator, BoolVal
from ..expressions.globalconstraints import GlobalConstraint
from ..expressions.variables import _BoolVarImpl, NegBoolView, _IntVarImpl, _NumVarImpl, intvar, boolvar
from ..expressions.utils import is_num, is_any_list, get_bounds
from ..transformations.get_variables import get_variables
from ..transformations.linearize import canonical_comparison
from ..transformations.normalize import toplevel_list
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.flatten_model import flatten_constraint
from ..transformations.comparison import only_numexpr_equality
from ..transformations.reification import reify_rewrite, only_bv_reifies, only_implies
from ..transformations.safening import no_partial_functions

import time


class CPM_pumpkin(SolverInterface):
    """
    Interface to Pumpkin's API

    Creates the following attributes (see parent constructor for more):

    - ``pum_solver``: the pumpkin.Model() object
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            import pumpkin_solver as psp
            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e


    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return version('pumpkin-solver')
        except PackageNotFoundError:
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
            cpm_model: Model(), a CPMpy Model() (optional)
            subsolver: None, not used
        """
        if not self.supported():
            raise Exception("CPM_Pumpkin: Install the python package 'pumpkin_solver'")

        from pumpkin_solver import Model

        assert subsolver is None 

        # initialise the native solver object
        self.pum_solver = Model() 
        self.predicate_map = {} # cache predicates for reuse

        # for objective
        self._objective = None
        self.objective_is_min = True

        # initialise everything else and post the constraints/objective
        super().__init__(name="pumpkin", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.pum_solver


    def solve(self, time_limit=None, prove=False, proof_name="proof.drcp", proof_location=".", assumptions=None):
        """
        Call the Pumpkin solver

        Arguments:
            time_limit (float, optional):  maximum solve time in seconds 
            prove: whether to produce a DRCP proof (.lits file and .drcp proof file).
            proof_name: name for the the proof files.
            proof_location: location for the proof files (default to current working directory).
            assumptions: CPMpy Boolean variables (or their negation) that are assumed to be true.
                            For repeated solving, and/or for use with :func:`s.get_core() <cpmpy.solvers.pumpkin.CPM_pumpkin.get_core>`: if the model is UNSAT,
                            `get_core()` returns a small subset of assumption variables that are unsat together.
        """

        # Again, I don't know why this is necessary, but the PyO3 modules seem to be a bit wonky.
        from pumpkin_solver import BoolExpression as PumpkinBool, IntExpression as PumpkinInt
        from pumpkin_solver import SatisfactionResult, SatisfactionUnderAssumptionsResult
        from pumpkin_solver.optimisation import OptimisationResult, Direction

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # parse and dispatch the arguments
        if time_limit is not None and time_limit < 0:
            raise ValueError("Time limit cannot be negative, but got {time_limit}")
        kwargs = dict(timeout=time_limit)

        if self.has_objective():
            assert assumptions is None, "Optimization under assumptions is not supported"
            solve_func = self.pum_solver.optimise
            kwargs.update(proof=join(proof_location, proof_name) if prove else None,
                          objective=self.solver_var(self._objective),
                          direction=Direction.Minimise if self.objective_is_min else Direction.Maximise)

        elif assumptions is not None:
            assert not prove, "Proof-logging under assumptions is not supported"
            pum_assumptions = [self.to_predicate(a, tag=None) for a in assumptions]
            self.assump_map = dict(zip(pum_assumptions, assumptions))
            solve_func = self.pum_solver.satisfy_under_assumptions
            kwargs.update(assumptions=pum_assumptions)

        else:
            solve_func = self.pum_solver.satisfy
            kwargs.update(proof=join(proof_location, proof_name) if prove else None)

        self._pum_core = None
        
        start_time = time.time() # when did solving start
        # call the solver, with parameters
        result = solve_func(**kwargs)

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = time.time() - start_time

        # translate solver exit status to CPMpy exit status
        if self.has_objective(): # check result after optimisation
            if isinstance(result, OptimisationResult.Optimal):
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            elif isinstance(result, OptimisationResult.Satisfiable):
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            elif  isinstance(result, OptimisationResult.Unsatisfiable):
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
            elif isinstance(result, OptimisationResult.Unknown):
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            else:
                raise ValueError(f"Unknown Pumpkin-result: {result} of type {type(result)}, please report on github...")

        elif assumptions is not None: # check result under assumptions
            if isinstance(result, SatisfactionUnderAssumptionsResult.Satisfiable):
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            elif isinstance(result, SatisfactionUnderAssumptionsResult.Unsatisfiable):
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
                self._pum_core = [] # empty core, no required assumptions to prove UNSAT
            elif isinstance(result, SatisfactionUnderAssumptionsResult.UnsatisfiableUnderAssumptions):
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
                self._pum_core = result._0
            elif isinstance(result, SatisfactionUnderAssumptionsResult.Unknown):
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            else:
                raise ValueError(f"Unknown Pumpkin-result: {result} of type {type(result)}, please report on github...")

        else: # satisfaction result without assumptions
            if isinstance(result, SatisfactionResult.Satisfiable):
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            elif isinstance(result, SatisfactionResult.Unsatisfiable):
                self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
            elif isinstance(result, SatisfactionResult.Unknown):
                self.cpm_status.exitstatus = ExitStatus.UNKNOWN
            else:
                raise ValueError(f"Unknown Pumpkin-result: {result} of type {type(result)}, please report on github...")

        # translate solution (of user vars only)
        has_sol = self._solve_return(self.cpm_status)
        if has_sol:
            solution = result._0

            # fill in variable values
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if isinstance(sol_var, PumpkinInt):
                    cpm_var._value = solution.int_value(sol_var)
                elif isinstance(sol_var, PumpkinBool):
                    cpm_var._value = solution.bool_value(sol_var)
                else:
                    raise NotSupportedError("Only boolean and integer variables are supported.")

            # translate solution values
            if self.has_objective():
                self.objective_value_ = solution.int_value(self.solver_var(self._objective))

        else: # wipe results
            for cpm_var in self.user_vars:
                cpm_var._value = None

        return has_sol


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """

        if is_num(cpm_var):
            return cpm_var

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            return self.solver_var(cpm_var._bv).negate()

        # create if it does not exist
        if isinstance(cpm_var, _NumVarImpl):
            if cpm_var not in self._varmap:
                if isinstance(cpm_var, _BoolVarImpl):
                    revar = self.pum_solver.new_boolean_variable(name=str(cpm_var))
                elif isinstance(cpm_var, _IntVarImpl):
                    revar = self.pum_solver.new_integer_variable(cpm_var.lb, cpm_var.ub, name=str(cpm_var))
                else:
                    raise NotImplementedError("Not a known var {}".format(cpm_var))
                self._varmap[cpm_var] = revar

            # return from cache
            return self._varmap[cpm_var]
        
        # can also be a scaled variable (multiplication view)
        elif isinstance(cpm_var, Operator) and cpm_var.name == "mul":
            const, cpm_var = cpm_var.args
            if not is_num(const):
                raise ValueError(f"Cannot create view from non-constant multiplier {const} * {cpm_var}")
            return self.solver_var(cpm_var).scaled(const)
        
        raise ValueError(f"Not a known var {cpm_var}")



    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize

            Arguments:
                expr: Expression, the CPMpy expression that represents the objective function
                minimize: Bool, whether it is a minimization problem (True) or maximization problem (False)

            'objective()' can be called multiple times, only the last one is stored

            .. note::
                technical side note: any constraints created during conversion of the objective
                are premanently posted to the solver
        """
        # make objective function non-nested
        obj_var = intvar(*get_bounds(expr))
        self += expr == obj_var

        # make objective function or variable and post
        self._objective = obj_var
        self.objective_is_min = minimize

    def has_objective(self):
        return self._objective is not None

    # `__add__()` first calls `transform()`
    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        # apply transformations
        cpm_cons = toplevel_list(cpm_expr)
        supported = {"alldifferent", "cumulative", "table", "negative_table", "InDomain"
                     "min", "max", "element", "abs"}
        cpm_cons = no_partial_functions(cpm_cons, safen_toplevel={"element"}) # safen toplevel elements, assume total decomposition for partial functions
        cpm_cons = decompose_in_tree(cpm_cons, supported=supported, csemap=self._csemap)
        cpm_cons = flatten_constraint(cpm_cons, csemap=self._csemap)  # flat normal form
        cpm_cons = only_bv_reifies(cpm_cons, csemap=self._csemap)
        cpm_cons = only_implies(cpm_cons, csemap=self._csemap)
        supported_halfreif = {"or", "sum", "wsum", "sub", "mul", "div", "abs", "min", "max"}
        cpm_cons = reify_rewrite(cpm_cons, supported=supported_halfreif, csemap=self._csemap) # reified element not supported yet
        cpm_cons = only_numexpr_equality(cpm_cons, supported=frozenset(["sum", "wsum", "sub"]),csemap=self._csemap)  # supports >, <, !=
        cpm_cons = canonical_comparison(cpm_cons) # ensure rhs is always a constant
        return cpm_cons

    def to_predicate(self, cpm_expr, tag=None):
        """
            Convert a CPMpy expression to a Pumpkin predicate (comparison with constant)
        """
        from pumpkin_solver import Comparator, Predicate

        if isinstance(cpm_expr, _BoolVarImpl):
            if isinstance(cpm_expr, NegBoolView):
                lhs, comp, rhs = cpm_expr._bv, Comparator.LessThanOrEqual, 0
            else:
                lhs, comp, rhs = cpm_expr, Comparator.GreaterThanOrEqual, 1

        elif isinstance(cpm_expr, Comparison):

            lhs, rhs = cpm_expr.args
            assert is_num(rhs), "rhs of comparison must be a constant to be a predicate"

            if isinstance(lhs, Operator): # can be sum with single arg
                if lhs.name == "sum" and len(lhs.args) == 1:
                    lhs = lhs.args[0]
                elif lhs.name == "wsum" and len(lhs.args[0]) == 1:
                    lhs = lhs.args[0][0] * lhs.args[1][0]
                elif lhs.name == "mul" and is_num(lhs.args[0]):
                    lhs = lhs.args[0] * lhs.args[1]
                else:
                    raise ValueError(f"Lhs of predicate should be a sum with 1 argument, wsum with 1 arg, or mul with const, but got {lhs}")

            if cpm_expr.name == "==": comp = Comparator.Equal
            if cpm_expr.name == "<=": comp = Comparator.LessThanOrEqual
            if cpm_expr.name == ">=": comp = Comparator.GreaterThanOrEqual
            if cpm_expr.name == "!=": comp = Comparator.NotEqual
            if cpm_expr.name == "<":  comp, rhs = Comparator.LessThanOrEqual, rhs - 1
            if cpm_expr.name == ">":  comp, rhs = Comparator.GreaterThanOrEqual, rhs + 1
        else:
            raise ValueError(f"Cannot convert CPMpy expression {cpm_expr} to a predicate")
        
        # do we already have this predicate? 
        #  (actually, cse might already catch these...)
        if (lhs, comp, rhs) not in self.predicate_map:
            if tag is None:
                tag = self.pum_solver.new_constraint_tag()
            pred = Predicate(self.to_pum_ivar(lhs, tag=tag), comp, rhs)
            self.predicate_map[(lhs, comp, rhs)] = pred
        
        return self.predicate_map[(lhs, comp, rhs)]



    def to_pum_ivar(self, cpm_var, tag=None):
        """
            Helper function to convert (boolean) variables and constants to Pumpkin integer expressions
        """
        if is_any_list(cpm_var):
            return [self.to_pum_ivar(v, tag=tag) for v in cpm_var]
        elif isinstance(cpm_var, _BoolVarImpl):
            return self.pum_solver.boolean_as_integer(self.solver_var(cpm_var), tag=tag)
        elif is_num(cpm_var):
            return self.solver_var(intvar(cpm_var, cpm_var))
        else:
            return self.solver_var(cpm_var)


    def _sum_args(self, expr, negate=False, tag=None):
        """
            Helper function to convert CPMpy sum-like operators into pumpkin-compatible arguments.
            expr is expected to be a `sum`, `wsum` or `sub` operator.

            :return: Returns a list of Pumpkin integer expressions
        """
        args = []
        if tag is None: raise ValueError("Expected tag to be provided but got None")
        if isinstance(expr, Operator) and expr.name == "sum":
            for cpm_var in expr.args:
                pum_var = self.solver_var(cpm_var)
                if cpm_var.is_bool(): # have convert to integer
                    pum_var = self.pum_solver.boolean_as_integer(pum_var, tag=tag)
                args.append(pum_var.scaled(-1 if negate else 1))
        elif isinstance(expr, Operator) and expr.name == "wsum":
            for w, cpm_var in zip(*expr.args):
                if w == 0: continue # exclude
                pum_var = self.solver_var(cpm_var)
                if cpm_var.is_bool(): # have convert to integer
                    pum_var = self.pum_solver.boolean_as_integer(pum_var, tag=tag)
                args.append(pum_var.scaled(-w if negate else w))
        elif isinstance(expr, Operator) and expr.name == "sub":
            x, y = self.solver_vars(expr.args)
            if expr.args[0].is_bool():
                x = self.pum_solver.boolean_as_integer(x, tag=tag)
            if expr.args[1].is_bool():
                y = self.pum_solver.boolean_as_integer(y, tag=tag)
            args = [x.scaled(-1 if negate else 1), y.scaled(1 if negate else -1)]
        else:
            raise ValueError(f"Unknown expression to convert in sum-arguments: {expr}")
        return args
    
    def _is_predicate(self, cpm_expr):
        """
            Check if a CPMpy expression can be converted to a Pumpkin predicate
        """
        if isinstance(cpm_expr, _NumVarImpl):
            return True
        elif isinstance(cpm_expr, Operator):
            if cpm_expr.name == "sum" and len(cpm_expr.args) == 1:
                return True
            if cpm_expr.name == "wsum" and len(cpm_expr.args[0]) == 1:
                return True
            if cpm_expr.name == "mul" and is_num(cpm_expr.args[0]):
                return True
        return False

    def _get_constraint(self, cpm_expr, tag=None):
        """
            Convert a CPMpy expression into a Pumpkin constraint
            Expects a transformed CPMpy expression, this logic is implemented as a separate function so we can support reification in `add()`
        """
        from pumpkin_solver import constraints
        if tag is None:
            tag = self.pum_solver.new_constraint_tag()
        if isinstance(cpm_expr, _BoolVarImpl):
            # base case, just var or ~var, post as clause
            return [constraints.Clause([self.solver_var(cpm_expr)], constraint_tag=tag)]
            
        elif isinstance(cpm_expr, Operator):
            if cpm_expr.name == "or": 
                return [constraints.Clause(self.solver_vars(cpm_expr.args), constraint_tag=tag)]

            raise NotImplementedError("Pumpkin: operator not (yet) supported", cpm_expr)

        elif isinstance(cpm_expr, Comparison):
            lhs, rhs = cpm_expr.args
            assert isinstance(lhs, Expression), f"Expected a CPMpy expression on lhs but got {lhs} of type {type(lhs)}"

            if self._is_predicate(lhs):
                pred = self.to_predicate(cpm_expr, tag=tag)
                return [constraints.Clause([self.pum_solver.predicate_as_boolean(pred, tag=tag)], constraint_tag=tag)]

            if cpm_expr.name == "==":
                
                if "sum" in lhs.name or lhs.name == "sub":
                    return [constraints.Equals(self._sum_args(lhs, tag=tag), rhs, constraint_tag=tag)]
               
                pum_rhs = self.to_pum_ivar(rhs, tag=tag) # other operators require IntExpression
                if lhs.name == "div":
                    return [constraints.Division(*self.to_pum_ivar(lhs.args, tag=tag), pum_rhs, constraint_tag=tag)]
                elif lhs.name == "mul":
                    return [constraints.Times(*self.to_pum_ivar(lhs.args, tag=tag), pum_rhs, constraint_tag=tag)]
                elif lhs.name == "abs":
                    return [constraints.Absolute(self.to_pum_ivar(lhs.args[0], tag=tag), pum_rhs, constraint_tag=tag)]
                elif lhs.name == "min":
                    return [constraints.Minimum(self.to_pum_ivar(lhs.args, tag=tag), pum_rhs, constraint_tag=tag)]
                elif lhs.name == "max":
                    return [constraints.Maximum(self.to_pum_ivar(lhs.args, tag=tag), pum_rhs, constraint_tag=tag)]
                elif lhs.name == "element":
                    arr, idx = lhs.args
                    return [constraints.Element(self.to_pum_ivar(idx, tag=tag),
                                                self.to_pum_ivar(arr, tag=tag),
                                                pum_rhs, constraint_tag=tag)]
                else:
                    raise NotImplementedError("Unknown lhs of comparison", cpm_expr)

            elif cpm_expr.name == "<=":
                return [constraints.LessThanOrEquals(self._sum_args(lhs, tag=tag), rhs, constraint_tag=tag)]
            elif cpm_expr.name == "<":
                return [constraints.LessThanOrEquals(self._sum_args(lhs, tag=tag), rhs-1,constraint_tag=tag)]
            elif cpm_expr.name == ">=":
                return [constraints.LessThanOrEquals(self._sum_args(lhs, negate=True, tag=tag), -rhs, constraint_tag=tag)]
            elif cpm_expr.name == ">":
                return [constraints.LessThanOrEquals(self._sum_args(lhs, negate=True, tag=tag), -rhs-1, constraint_tag=tag)]
            elif cpm_expr.name == "!=":
                return [constraints.NotEquals(self._sum_args(lhs, tag=tag), rhs, constraint_tag=tag)]

            raise ValueError("Unknown comparison", cpm_expr)

        elif isinstance(cpm_expr, GlobalConstraint):
            if cpm_expr.name == "alldifferent":
                return [constraints.AllDifferent(self.solver_vars(cpm_expr.args), constraint_tag=tag)]
            
            elif cpm_expr.name == "cumulative":
                start, dur, end, demand, cap = cpm_expr.args
                assert all(is_num(d) for d in dur), "Pumpkin only accepts Cumulative with fixed durations"
                assert all(is_num(d) for d in demand), "Pumpkin only accepts Cumulative with fixed demand"
                assert is_num(cap), "Pumpkin only accepts Cumulative with fixed capacity"

                dur_cons = []
                for c in self.transform([s + d == e for s,d,e in zip(start, dur, end)]):
                    dur_cons += self._get_constraint(c, tag=tag)

                return [constraints.Cumulative(self.solver_vars(start),dur, demand, cap, constraint_tag=tag)] + dur_cons

            elif cpm_expr.name == "table":
                arr, table = cpm_expr.args
                return [constraints.Table(self.to_pum_ivar(arr), 
                                          np.array(table).tolist(), # ensure Python list
                                          constraint_tag=tag)
                        ]
            
            elif cpm_expr.name == "negative_table":
                arr, table = cpm_expr.args
                return [constraints.NegativeTable(self.to_pum_ivar(arr), 
                                                  np.array(table).tolist(),# ensure Python list
                                                  constraint_tag=tag)  
                        ]
            
            elif cpm_expr.name == "InDomain":
                val, domain = cpm_expr.args
                return [constraints.Table([self.to_pum_ivar(val)], 
                                          np.array(domain).tolist(), # ensure Python list
                                          constraint_tag=tag)
                        ] 
            
            
            else:
                raise NotImplementedError(f"Unknown global constraint {cpm_expr}")
                

        elif isinstance(cpm_expr, BoolVal): # unlikely base case
            a = boolvar() # dummy variable
            if cpm_expr.value() is True:
                return self._get_constraint(Operator("sum", [a]) >= 1)
            else:
                return self._get_constraint(Operator("sum", [a]) <= -1)

        else:
            raise ValueError("Unexpected constraint:", cpm_expr)


    def add(self, cpm_expr_orig):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: self
        """

        # add new user vars to the set
        get_variables(cpm_expr_orig, collect=self.user_vars)

        for cpm_expr in self.transform(cpm_expr_orig):
            if isinstance(cpm_expr, Operator) and cpm_expr.name == "->": # found implication
                bv, subexpr = cpm_expr.args
                for pum_cons in self._get_constraint(subexpr):
                    self.pum_solver.add_implication(pum_cons, self.solver_var(bv))
            else:
                for pum_cons in self._get_constraint(cpm_expr):
                    self.pum_solver.add_constraint(pum_cons)

        return self
    __add__ = add # avoid redirect in superclass


    def get_core(self):
        """
           For use with :func:`s.solve(assumptions=[...]) <cpmpy.solvers.pumpkin.CPM_pumpkin.solve>`. Only meaningful if the solver returned UNSAT. In that case, get_core() returns a small subset of assumption variables that are unsat together.

           CPMpy will return only those variables that are False (in the UNSAT core)

            .. note::

                Note that there is no guarantee that the core is minimal, though this interface does open up the possibility to add more advanced Minimal Unsatisfiabile Subset algorithms on top. All contributions welcome!
        """

        assert self._pum_core is not None, "Can only get core if the last solve-call was unsatisfiable under assumptions"
        return [self.assump_map[pred] for pred in self._pum_core]




#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## utils.py
##
"""
    Utilities for handling solvers

    =================
    List of functions
    =================

    .. autosummary::
        :nosignatures:

        param_combinations
"""

import warnings # for deprecation warning

from .gurobi import CPM_gurobi
from .ortools import CPM_ortools
from .minizinc import CPM_minizinc
from .pysat import CPM_pysat
from .z3 import CPM_z3
from .gcs import CPM_gcs
from .pysdd import CPM_pysdd
from .exact import CPM_exact
from .choco import CPM_choco
from .pumpkin import CPM_pumpkin
from .cpo   import CPM_cpo
from .cplex import CPM_cplex
from .pindakaas import CPM_pindakaas
from .hexaly import CPM_hexaly

def param_combinations(all_params, remaining_keys=None, cur_params=None):
    """
        Recursively yield all combinations of param values

        For example usage, see `examples/advanced/hyperparameter_search.py`
        https://github.com/CPMpy/cpmpy/blob/master/examples/advanced/hyperparameter_search.py

        - all_params is a dict of `{key: list}` items, e.g.:
          ``{'val': [1,2], 'opt': [True,False]}``

        - output is an generator over all `{key:value}` combinations
          of the keys and values. For the example above:
          ``generator([{'val':1,'opt':True},{'val':1,'opt':False},{'val':2,'opt':True},{'val':2,'opt':False}])``
    """
    if remaining_keys is None or cur_params is None:
        # init
        remaining_keys = list(all_params.keys())
        cur_params = dict()

    cur_key = remaining_keys[0]
    for cur_value in all_params[cur_key]:
        cur_params[cur_key] = cur_value
        if len(remaining_keys) == 1:
            # terminal, return copy
            yield dict(cur_params)
        else:
            # recursive call
            yield from param_combinations(all_params, 
                            remaining_keys=remaining_keys[1:],
                            cur_params=cur_params)

class SolverLookup():
    @classmethod
    def base_solvers(cls):
        """
            Return ordered list of (name, class) of base CPMpy
            solvers

            First one is default
        """
        return [
                ("ortools", CPM_ortools),
                ("z3", CPM_z3),
                ("minizinc", CPM_minizinc),
                ("gcs", CPM_gcs),
                ("gurobi", CPM_gurobi),
                ("pysat", CPM_pysat),
                ("pysdd", CPM_pysdd),
                ("exact", CPM_exact),
                ("choco", CPM_choco),
                ("pumpkin", CPM_pumpkin),
                ("cpo", CPM_cpo),
                ("cplex", CPM_cplex),
                ("pindakaas", CPM_pindakaas),
                ("hexaly", CPM_hexaly)
               ]

    @classmethod
    def print_status(cls):
        """
            Print all CPMpy solvers and their installation status on this system.
        """
        for (basename, CPM_slv) in cls.base_solvers():
            if CPM_slv.supported():
                print(f"{basename}: Supported, ready to use.")
            else:
                print(f"{basename}: Not supported (missing Python package, binary or license).")

    @classmethod
    def supported(cls):
        """
            Return the list of names of all solvers (and subsolvers) supported on this system.

            If a solver name is returned, it means that the solver's `.supported()` function returns True
            and it is hence ready for immediate use
            (e.g. any separate binaries are also installed if necessary, and licenses are active if needed).

            Typical use case is to use these names in `SolverLookup.get(name)`.
        """
        names = []
        for (basename, CPM_slv) in cls.base_solvers():
            if CPM_slv.supported():
                names.append(basename)
                if hasattr(CPM_slv, "solvernames"):
                    subnames = CPM_slv.solvernames(installed=True)
                    for subn in subnames:
                        names.append(basename+":"+subn)
        return names

    @classmethod
    def solvernames(cls):
        # The older (more indirectly named) way to get the list of names of *supported* solvers.
        # Will be deprecated at some point.
        return cls.supported()

    @classmethod
    def get(cls, name=None, model=None, **init_kwargs):
        """
            get a specific solver (by name), with 'model' passed to its constructor

            This is the preferred way to initialise a solver from its name

            :param name: name of the solver to use
            :param model: model to pass to the solver constructor
            :param init_kwargs: additional keyword arguments to pass to the solver constructor
        """
        solver_cls = cls.lookup(name=name)

        # check for a 'solver:subsolver' name
        subname = None
        if name is not None and ':' in name:
            _,subname = name.split(':',maxsplit=1)
        return solver_cls(model, subsolver=subname, **init_kwargs)

    @classmethod
    def lookup(cls, name=None):
        """
            lookup a solver _class_ by its name

            warning: returns a 'class', not an object!
            see get() for normal uses
        """
        if name is None:
            # first solver class
            return cls.base_solvers()[0][1]

        # split name if relevant
        solvername = name
        subname = None
        if ':' in solvername:
            solvername,_ = solvername.split(':',maxsplit=1)

        for (basename, CPM_slv) in cls.base_solvers():
            if basename == solvername:
                # found the right solver
                return CPM_slv
        raise ValueError(f"Unknown solver '{name}', choose from {cls.solvernames()}")
 

    @classmethod
    def version(cls):
        """
        Returns an overview of all solvers supported by CPMpy as a list of dicts.

        Each dict consists of:

        - "name": <base_solver> or <base_solver>:<subsolver>
        - "installed": install status (True/False)
        - "version": version of solver's Python library (or one of its subsolvers if applicable)
        """
        result = []
        for (basename, CPM_slv) in cls.base_solvers():
            installed = CPM_slv.supported()
            version = CPM_slv.version() if installed and hasattr(CPM_slv, 'version') else None
            
            # Collect main solver status
            result.append({
                    "name": basename,
                    "installed": installed, 
                    "version": version,
                })
            
            # Handle subsolvers if applicable
            if installed and hasattr(CPM_slv, 'solvernames'):
                subnames = CPM_slv.solvernames()
                installed_subnames = CPM_slv.solvernames(installed=True)
                for subn in subnames:
                    is_installed = subn in installed_subnames
                    subsolver_status = {
                        "name": basename + ":" + subn, 
                        "installed": is_installed, 
                        "version": CPM_slv.solverversion(subn) if installed else None,
                    }
                    result.append(subsolver_status)  # Append subsolver status
        return result


    @classmethod
    def print_version(cls):
        """
        Prints a tabulated report on the different solvers supported by CPMpy,
        i.e. whether they are installed on the current system and if so which version.
        """
        
        # Get the solver information using the version() method
        solver_versions = cls.version()

        # Print the header
        print(f"{'Solver':<25} {'Installed':<10} {'Version':<15}")
        print("-" * 50)

        # Iterate over the solvers
        for solver_version in solver_versions:
            basename, installed, version = solver_version["name"], solver_version["installed"], solver_version["version"]

            # If this is a subsolver (indicated by a ':' in the name), indent the output
            if ':' in basename:
                print(f" ↪ {basename.split(':')[-1]:<22} {'Yes' if installed else 'No':<10} {(version if version else ' '):<15}")  # Subsolver with indentation
            else:
                # For main solvers, show version if available
                version = version if version else "Not found" if installed else "-"
                print(f"{basename:<25} {'Yes' if installed else 'No':<10} {version:<15}")


# using `builtin_solvers` is DEPRECATED, use `SolverLookup` object instead
# Order matters! first is default, then tries second, etc...
builtin_solvers = [CPM_ortools, CPM_gurobi, CPM_minizinc, CPM_pysat, CPM_exact, CPM_choco]
def get_supported_solvers():
    """
        Returns a list of solvers supported on this machine.
       
        .. deprecated:: 0.9.4
            Please use :class:`SolverLookup` object instead.

        :return: a list of SolverInterface sub-classes :list[SolverInterface]:
    """
    warnings.warn("Deprecated, use Model.solvernames() instead, will be removed in stable version", DeprecationWarning)
    return [sv for sv in builtin_solvers if sv.supported()]



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## python_builtins.py
##
"""
    Overwrites a number of python built-ins, so that they work over variables as expected.

    =================
    List of functions
    =================
    .. autosummary::
        :nosignatures:

        all
        any
        max
        min
        sum
        abs
"""
import builtins  # to use the original Python-builtins

from .utils import is_false_cst, is_true_cst, is_any_list
from .variables import NDVarArray, cpm_array
from .core import Expression, Operator, BoolVal
from .globalfunctions import Minimum, Maximum, Abs


# Overwriting all/any python built-ins
# all: listwise 'and'
def all(iterable):
    """
        all() overwrites python built-in,
        if iterable contains any `Expression`, then returns an Operator("and", iterable)
        otherwise returns whether all of the arguments are true
    """
    if isinstance(iterable, NDVarArray): iterable=iterable.flat # 1D iterator
    collect = [] # logical expressions
    is_expr, return_false = False, False
    for elem in iterable:
        if isinstance(elem, Expression): # probably most likely case
            is_expr = True
            if isinstance(elem, BoolVal):
                if not elem.args[0]: # False constant
                    return_false = True
            elif elem.is_bool():
                collect.append(elem)
            else:
                raise Exception("Non-Boolean argument '{}' to 'all'".format(elem))
        elif is_true_cst(elem):
            pass
        elif is_false_cst(elem):
            return_false = True
        elif isinstance(elem, list):
            raise Exception("Encountered list in 'all', only accept non-nested lists")
        else:
            raise Exception("Unexpected argument '{}' to 'all'".format(elem))
    
    if return_false:
        return BoolVal(False) if is_expr else False
    if len(collect) == 0:
        return BoolVal(True) if is_expr else True
    if len(collect) == 1:
        return collect[0]
    if len(collect) >= 2:
        return Operator("and", collect)
    raise Exception(f"Unepxected collection {collect}")


# any: listwise 'or'
def any(iterable):
    """
        any() overwrites python built-in,
        if iterable contains an `Expression`, then returns an Operator("or", iterable)
        otherwise returns whether any of the arguments is true
    """
    if isinstance(iterable, NDVarArray): iterable=iterable.flat # 1D iterator
    collect = [] # logical expressions
    is_expr, return_true = False, False
    for elem in iterable:
        if isinstance(elem, Expression): # probably most likely case
            is_expr = True
            if isinstance(elem, BoolVal):
                if elem.args[0]: # True constant
                    return_true = True
            elif elem.is_bool():
                collect.append(elem)
            else:
                raise Exception("Non-Boolean argument '{}' to 'all'".format(elem))
        elif is_true_cst(elem):
            return_true = True
        elif is_false_cst(elem):
            pass
        elif isinstance(elem, list):
            raise Exception("Encountered list in 'all', only accept non-nested lists")
        else:
            raise Exception("Unexpected argument '{}' to 'all'".format(elem))
    
    if return_true:
        return BoolVal(True) if is_expr else True
    if len(collect) == 0:
        return BoolVal(False) if is_expr else False
    if len(collect) == 1:
        return collect[0]
    if len(collect) >= 2:
        return Operator("or", collect)
    raise Exception(f"Unepxected collection {collect}")


def max(*iterable, **kwargs):
    """
        max() overwrites the python built-in to support decision variables.

        if iterable does not contain CPMpy expressions, the built-in is called
        else a Maximum functional global constraint is constructed; no keyword
        arguments are supported in that case
    """
    if len(iterable) == 1:
        iterable = tuple(iterable[0])
    if not builtins.any(isinstance(elem, Expression) for elem in iterable):
        return builtins.max(iterable, **kwargs)

    assert len(kwargs)==0, "max over decision variables does not support keyword arguments"
    return Maximum(iterable)


def min(*iterable, **kwargs):
    """
        min() overwrites the python built-in to support decision variables.

        if iterable does not contain CPMpy expressions, the built-in is called
        else a Minimum functional global constraint is constructed; no keyword
        arguments are supported in that case
    """
    if len(iterable) == 1:
        iterable = tuple(iterable[0])
    if not builtins.any(isinstance(elem, Expression) for elem in iterable):
        return builtins.min(iterable, **kwargs)

    assert len(kwargs)==0, "min over decision variables does not support keyword arguments"
    return Minimum(iterable)


def sum(*iterable, **kwargs):
    """
        sum() overwrites the python built-in to support decision variables.

        if iterable does not contain CPMpy expressions, the built-in is called
        checks if all constants and uses built-in sum() in that case
    """
    if len(iterable) == 1:
        iterable = tuple(iterable[0]) # Fix generator polling
    if not builtins.any(isinstance(elem, Expression) for elem in iterable):
        return builtins.sum(iterable, **kwargs)

    assert len(kwargs)==0, "sum over decision variables does not support keyword arguments"
    return Operator("sum", iterable)


def abs(element):
    """
        abs() overwrites the python built-in to support decision variables.

        if the element given is not a CPMpy expression, the built-in is called
        else an Absolute functional global constraint is constructed.
    """
    if is_any_list(element):  # compat: not allowed by builtins.abs(), but allowed by numpy.abs()
        return cpm_array([abs(elem) for elem in element])

    if isinstance(element, Expression):
        # create global
        return Abs(element)
    
    return builtins.abs(element)

    



"""
    Classes and functions that represent and create expressions (constraints and objectives)

    ==================
    List of submodules
    ==================
    .. autosummary::
        :nosignatures:

        variables
        core
        globalconstraints
        globalfunctions
        python_builtins
        utils


"""

# we only import methods/classes that are used for modelling
# others need to be imported by the developer explicitely
from .variables import boolvar, intvar, cpm_array
from .variables import BoolVar, IntVar, cparray # Old, to be deprecated
from .globalconstraints import AllDifferent, AllDifferentExcept0, AllDifferentExceptN, AllEqual, AllEqualExceptN, Circuit, Inverse, Table, ShortTable, Xor, Cumulative, \
    IfThenElse, GlobalCardinalityCount, DirectConstraint, InDomain, Increasing, Decreasing, IncreasingStrict, DecreasingStrict, \
    LexLess, LexLessEq, LexChainLess, LexChainLessEq, Precedence, NoOverlap, \
    NegativeTable, Regular
from .globalconstraints import alldifferent, allequal, circuit # Old, to be deprecated
from .globalfunctions import Maximum, Minimum, Abs, Element, Count, NValue, NValueExcept, Among
from .core import BoolVal
from .python_builtins import all, any, max, min, sum, abs



"""
Collection of visualisation tools for processing the result of a `benchmark.py` run.

Best used though its CLI, a command-line tool to visualize and analyze solver performance 
based on CSV output files.

E.g. to compare the results of multiple solvers:

.. code-block:: console

    python analyze.py <results_dir>

Positional Arguments
--------------------
files : str
    One or more CSV files (or a single directory) containing performance data to analyze.

Optional Arguments
------------------
--time_limit : float, optional
    Maximum time limit (in seconds) to display on the x-axis of the plot.

--output, -o : str, optional
    Path to save the generated plot image (e.g., "output.png"). If not provided, the plot will be displayed interactively.
"""

import argparse
import ast
from pathlib import Path
import re
import matplotlib
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

def _extract_cost(solution_str):
    """
    Extract numeric cost from solution string like '<instantiation ... cost="69">'
    """
    if isinstance(solution_str, str):
        match = re.search(r'cost="(\d+)"', solution_str)
        if match:
            return float(match.group(1))
    return np.nan


def xcsp3_plot(df, time_limit=None):
    # Get unique solvers
    solvers = df['solver'].unique()

    # Determine the status to plot (Opt if at least one opt, otherwise sat)
    statuses = df['status'].unique()
    if 'OPTIMUM FOUND' in statuses:
        status_filter = 'OPTIMUM FOUND'
    else:
        status_filter = 'SATISFIABLE'
    df = df[(df['status'] == status_filter) | (df['status'] == 'UNSATISFIABLE')]  # only those that reached the desired status

    # Count how many instances each solver solved (with correct status)
    solver_counts = df['solver'].value_counts()

    # Sort solvers descending by number of instances solved
    solvers_sorted = solver_counts.sort_values(ascending=False).index.tolist()
    
    # Create figure
    fig = plt.figure(figsize=(10, 6))
    
    for solver in sorted(solvers): # Sort solver names for consistent ordering
        # Get data for this solver
        solver_data = df[df['solver'] == solver]
        
        # Sort by time_total
        solver_data = solver_data.sort_values('time_total')
        
        # If time_limit is set, truncate data
        if time_limit is not None:
            solver_data = solver_data[solver_data['time_total'] <= time_limit]
        
        # Build x and y values
        x = [0.0] + solver_data['time_total'].tolist()
        y = [0] + list(range(1, len(solver_data) + 1))
        
        # Plot the performance curve
        plt.plot(x, y, label=f"{solver} ({len(solver_data)})", linewidth=2.5)
    
    # Set plot properties
    plt.xlabel('Time (seconds)')
    plt.ylabel(f'Number of instances returning \'{status_filter}\'')
    # Get unique year-track combinations
    year_track_pairs = df[['year', 'track']].drop_duplicates()
    datasets = ', '.join([f'{row.year}:{row.track}' for _, row in year_track_pairs.iterrows()])
    plt.title(f'Performance Plot ({datasets})')
    plt.grid(True)
    plt.legend()
    
    # Set x-axis limit if specified
    if time_limit is not None:
        plt.xlim(0, time_limit)

    return fig

def get_cost(row):
    """
    Get the achieved objective value from the provided row.
    If intermediate solutions are available, get the best found (not neccesarily proven optimal).
    """
    intermediate = row['intermediate']

    # Try to parse string representations safely
    if isinstance(intermediate, str):
        try:
            intermediate = ast.literal_eval(intermediate)
        except (ValueError, SyntaxError):
            intermediate = None

    # If it's a valid list of tuples, return the last objective
    if isinstance(intermediate, list) and len(intermediate) > 0:
        try:
            return intermediate[-1][1]
        except (IndexError, TypeError):
            pass

    # Fallback to extracting from solution
    return _extract_cost(row['solution'])

def xcsp3_objective_performance_profile(df):
    # Parse cost from the solution string
    df = df.copy()
    # print(df["intermediate"])
    df['cost'] = df.apply(get_cost, axis=1)
    # df['cost'] = df['solution'].apply(extract_cost)

    # Pivot to get costs per instance per solver
    pivot = df.pivot_table(index='instance', columns='solver', values='cost')

    # Drop instances not solved by all solvers (for fair comparison)
    pivot = pivot.dropna(how='all')

    # Compute the best (minimum) cost per instance
    best_costs = pivot.min(axis=1)

    # Compute performance ratios: solver_cost / best_cost
    perf_ratios = pivot.divide(best_costs, axis=0)

    # Replace inf or NaN with a large number for safe plotting
    perf_ratios = perf_ratios.replace([np.inf, np.nan], np.max(perf_ratios.values) * 10)

    # Compute a score for sorting: fraction of instances with ratio ≤ 1.1 (or similar)
    score_threshold = 1.1
    solver_scores = (perf_ratios <= score_threshold).mean().sort_values(ascending=False)
    sorted_solvers = solver_scores.index.tolist()

    # τ range for plotting
    tau_vals = np.linspace(1, perf_ratios.max().max(), 500)

    # Plotting
    fig = plt.figure(figsize=(10, 6))

    for solver in sorted_solvers:
        y_vals = [(perf_ratios[solver] <= tau).mean() for tau in tau_vals]
        plt.plot(tau_vals, y_vals, label=solver, linewidth=2.5)

    plt.xlabel(r'Objective ratio $\tau$')
    plt.ylabel('Fraction of instances')
    year_track_pairs = df[['year', 'track']].drop_duplicates()
    datasets = ', '.join([f'{row.year}:{row.track}' for _, row in year_track_pairs.iterrows()])
    plt.title(f'Objective Performance Profile ({datasets})')
    plt.grid(True)
    plt.legend()
    plt.xlim(left=1)

    return fig

def xcsp3_stats(df):

    for phase in ['parse', 'model', 'post']:
        slowest_idx = df[f'time_{phase}'].idxmax()
        if slowest_idx is not None and not pd.isna(slowest_idx):
            print(f"Slowest {phase}: {df.loc[slowest_idx, f'time_{phase}']}s ({df.loc[slowest_idx, 'instance']}, {df.loc[slowest_idx, 'solver']})")

    for solver in df['solver'].unique():
        solver_total = df[df['solver'] == solver]['time_total'].sum()
        print(f"Grand total for {solver}: {solver_total/60:.2f} minutes")
    
    
def main():
    # Set up argument parser
    parser = argparse.ArgumentParser(description='Analyze XCSP3 solver performance data')
    parser.add_argument('files', nargs='+', help='List of CSV files or directories to analyze')
    parser.add_argument('--time_limit', type=float, default=None, 
                        help='Maximum time limit in seconds to show on x-axis')
    parser.add_argument('--output', '-o', type=str, default=None,
                        help='Path to save the plot image (e.g., output.png)')
    args = parser.parse_args()
    
    # Gather all CSV files
    csv_files = []
    for path_str in args.files:
        path = Path(path_str)
        if path.is_file() and path.suffix == '.csv':
            csv_files.append(path)
        elif path.is_dir():
            csv_files.extend(path.rglob('*.csv'))
        else:
            print(f"Warning: {path} is not a valid CSV file or directory")

    if not csv_files:
        print("No CSV files found.")
        return

    # Read and merge all CSV files
    dfs = []
    for file in csv_files:
        df = pd.read_csv(file)
        dfs.append(df)
    
    merged_df = pd.concat(dfs, ignore_index=True)
    
    # Print some stats
    xcsp3_stats(merged_df)
    
    # Create performance plot
    fig = xcsp3_plot(merged_df, args.time_limit)
    # fig = xcsp3_objective_performance_profile(merged_df)

    # Save or show plot
    if args.output:
        fig.savefig(args.output, bbox_inches='tight')
        print(f"Plot saved to {args.output}")
    else:
        plt.show()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## expressions.py
##
"""
    The :class:`~cpmpy.expressions.core.Expression` superclass and common subclasses :class:`~cpmpy.expressions.core.Comparison` and :class:`~cpmpy.expressions.core.Operator`.
    
    None of these objects should be directly created, they are automatically created through operator
    overloading on variables and expressions.

    Here is a list of standard python operators and what object (with what expr.name) it creates:

    Comparisons
    -----------
    ===================  ==========================
    Python Operator      CPMpy Object                     
    ===================  ==========================
    `x == y`             `Comparison("==", x, y)`         
    `x != y`             `Comparison("!=", x, y)`         
    `x < y`              `Comparison("<", x, y)`          
    `x <= y`             `Comparison("<=", x, y)`         
    `x > y`              `Comparison(">", x, y)`          
    `x >= y`             `Comparison(">=", x, y)`         
    ===================  ==========================

    Arithmetic Operators
    --------------------
    ===========================  ===============================================
    Python Operator              CPMpy Object                                  
    ===========================  ===============================================
    `-x`                         `Operator("-", [x])`                          
    `x + y`                      `Operator("sum", [x, y])`                     
    `sum([x,y,z])`               `Operator("sum", [x, y, z])`                  
    `sum([c0*x, c1*y, c2*z])`    `Operator("wsum", [[c0, c1, c2], [x, y, z]])` 
    `x - y`                      `Operator("sum", [x, -y])`                    
    `x * y`                      `Operator("mul", [x, y])`                     
    `x // y`                     `Operator("div", [x, y])` (integer division)
    `x % y`                      `Operator("mod", [x, y])` (modulo)                    
    `x ** y`                     `Operator("pow", [x, y])` (power)  
    ===========================  ===============================================                   

    
    Logical Operators
    -----------------
    ===================  =======================================================
    Python Operator      CPMpy Object  
    ===================  =======================================================                                        
    `x & y`              `Operator("and", [x, y])`                             
    `x | y`              `Operator("or", [x, y])`                              
    `~x`                 `Operator("not", [x])` or `NegBoolView(x)` if Boolean 
    `x ^ y`              `Xor([x, y])`                                         
    ===================  =======================================================

    Python has no built-in operator for `implication` that can be overloaded.
    CPMpy hence has a function :func:`~cpmpy.expressions.core.Expression.implies` that can be called:

    ===================  ======================
    Python Operator      CPMpy Object          
    ===================  ======================
    `x.implies(y)`       `Operator("->", [x,y])` 
    ===================  ======================

    Apart from operator overloading, expressions implement two important functions:

    - :func:`~cpmpy.expressions.core.Expression.is_bool`   
        which returns whether the return type of the expression is Boolean.
        If it does, the expression can be used as top-level constraint
        or in logical operators.

    - :func:`~cpmpy.expressions.core.Expression.value`    
        computes the value of this expression, by calling .value() on its
        subexpressions and doing the appropriate computation
        this is used to conveniently print variable values, objective values
        and any other expression value (e.g. during debugging).
    
    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        Expression
        Comparison
        Operator
"""
import copy
import warnings
from types import GeneratorType
import numpy as np
import cpmpy as cp

from .utils import is_int, is_num, is_any_list, flatlist, get_bounds, is_boolexpr, is_true_cst, is_false_cst, argvals, is_bool
from ..exceptions import IncompleteFunctionError, TypeError


class Expression(object):
    """
    An Expression represents a symbolic function with a `self.name` and `self.args` (arguments)

    Each Expression is considered to be a function whose value can be used
      in other expressions

    Expressions may implement:

    - :func:`~cpmpy.expressions.core.Expression.is_bool`:               whether its return type is Boolean
    - :func:`~cpmpy.expressions.core.Expression.value`:                 the value of the expression, default None
    - :func:`implies(x) <cpmpy.expressions.core.Expression.implies>`:   logical implication of this expression towards `x`
    - :func:`~cpmpy.expressions.core.Expression.__repr__`:              for pretty printing the expression
    - any ``__op__`` python operator overloading
    """

    def __init__(self, name, arg_list):
        self.name = name

        if isinstance(arg_list, (tuple, GeneratorType)):
            arg_list = list(arg_list)
        elif isinstance(arg_list, np.ndarray):
            # must flatten
            arg_list = arg_list.reshape(-1)
        for i in range(len(arg_list)):
            if isinstance(arg_list[i], np.ndarray):
                # must flatten
                arg_list[i] = arg_list[i].reshape(-1)

        assert (is_any_list(arg_list)), "_list_ of arguments required, even if of length one e.g. [arg]"
        self._args = arg_list


    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, args):
        raise AttributeError("Cannot modify read-only attribute 'args', use 'update_args()'")

    def update_args(self, args):
        """ Allows in-place update of the expression's arguments.
            Resets all cached computations which depend on the expression tree.
        """
        self._args = args
        # Reset cached "_has_subexpr"
        if hasattr(self, "_has_subexpr"):
            del self._has_subexpr

    def set_description(self, txt, override_print=True, full_print=False):
        self.desc = txt
        self._override_print = override_print
        self._full_print = full_print

    def __str__(self):
        if not hasattr(self, "desc") or self._override_print is False:
            return self.__repr__()
        out = self.desc
        if self._full_print:
            out += " -- "+self.__repr__()
        return out


    def __repr__(self):
        strargs = []
        for arg in self.args:
            if isinstance(arg, np.ndarray):
                # flatten
                strarg = ",".join(map(str, arg.flat))
                strargs.append(f"[{strarg}]")
            else:
                strargs.append(f"{arg}")
        return "{}({})".format(self.name, ",".join(strargs))

    def __hash__(self):
        return hash(self.__repr__())

    def has_subexpr(self):
        """ Does it contains nested :class:`Expressions <cpmpy.expressions.core.Expression>` (anything other than a :class:`~cpmpy.expressions.variables._NumVarImpl` or a constant)?
            Is of importance when deciding whether certain transformations are needed
            along particular paths of the expression tree.
            Results are cached for future calls and reset when the expression changes
            (in-place argument update).
        """
        # return cached result
        if hasattr(self, '_has_subexpr'):
            return self._has_subexpr

        # Initialize stack with args
        stack = list(self.args)

        while stack:
            el = stack.pop()
            if isinstance(el, Expression):
                # only 3 types of expressions are leafs: _NumVarImpl, BoolVal or NDVarArray with no expressions inside.
                if isinstance(el, cp.variables.NDVarArray) and el.has_subexpr():
                    self._has_subexpr = True
                    return True
                elif not isinstance(el, (cp.variables._NumVarImpl, BoolVal)):
                    self._has_subexpr = True
                    return True
            elif is_any_list(el):
                # Add list elements to stack for processing
                stack.extend(el)

        # No subexpressions found
        self._has_subexpr = False
        return False

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
            Default: yes
        """
        return True

    def value(self):
        return None # default

    def get_bounds(self):
        if self.is_bool():
            return 0, 1 #default for boolean expressions
        raise NotImplementedError(f"`get_bounds` is not implemented for type {self}")

    # keep for backwards compatibility
    def deepcopy(self, memodict={}):
        warnings.warn("Deprecated, use copy.deepcopy() instead, will be removed in stable version", DeprecationWarning)
        return copy.deepcopy(self, memodict)

    # implication constraint: self -> other
    # Python does not offer relevant syntax...
    # for double implication, use equivalence self == other
    def implies(self, other):
        # other constant
        if is_true_cst(other):
            return BoolVal(True)
        if is_false_cst(other):
            return ~self
        return Operator('->', [self, other])

    # Comparisons
    def __eq__(self, other):
        # BoolExpr == 1|true|0|false, common case, simply BoolExpr
        if self.is_bool() and is_num(other):
            if other is True or other == 1:
                return self
            if other is False or other == 0:
                return ~self
        return Comparison("==", self, other)

    def __ne__(self, other):
        return Comparison("!=", self, other)

    def __lt__(self, other):
        return Comparison("<", self, other)

    def __le__(self, other):
        return Comparison("<=", self, other)

    def __gt__(self, other):
        return Comparison(">", self, other)

    def __ge__(self, other):
        return Comparison(">=", self, other)

    # Boolean Operators
    # Implements bitwise operations & | ^ and ~ (and, or, xor, not)
    def __and__(self, other):
        # some simple constant removal
        if is_true_cst(other):
            return self
        # catch beginner mistake
        if is_num(other) and not is_bool(other):
            raise TypeError(f"{self}&{other} is not valid because {other} is a number, did you forget to put brackets? "
                            f"E.g. always write (x==2)&(y<5).")
        return Operator("and", [self, other])

    def __rand__(self, other):
        # some simple constant removal
        if is_true_cst(other):
            return self
        # catch beginner mistake
        if is_num(other) and not is_bool(other):
            raise TypeError(f"{other}&{self} is not valid because {other} is a number, "
                            f"did you forget to put brackets? E.g. always write (x==2)&(y<5).")
        return Operator("and", [other, self])

    def __or__(self, other):
        # some simple constant removal
        if is_false_cst(other):
            return self
        # catch beginner mistake
        if is_num(other) and not is_bool(other):
            raise TypeError(f"{self}|{other} is not valid because {other} is a number, "
                            f"did you forget to put brackets? E.g. always write (x==2)|(y<5).")
        return Operator("or", [self, other])

    def __ror__(self, other):
        # some simple constant removal
        if is_false_cst(other):
            return self
        # catch beginner mistake
        if is_num(other) and not is_bool(other):
            raise TypeError(f"{other}|{self} is not valid because {other} is a number, "
                            f"did you forget to put brackets? E.g. always write (x==2)|(y<5).")
        return Operator("or", [other, self])

    def __xor__(self, other):
        # some simple constant removal
        if is_true_cst(other):
            return ~self
        if is_false_cst(other):
            return self
        return cp.Xor([self, other])

    def __rxor__(self, other):
        # some simple constant removal
        if is_true_cst(other):
            return ~self
        if is_false_cst(other):
            return self
        return cp.Xor([other, self])

    # Mathematical Operators, including 'r'everse if it exists
    # Addition
    def __add__(self, other):
        if is_num(other) and other == 0:
            return self
        return Operator("sum", [self, other])

    def __radd__(self, other):
        if is_num(other) and other == 0:
            return self
        return Operator("sum", [other, self])

    # substraction
    def __sub__(self, other):
        # if is_num(other) and other == 0:
        #     return self
        # return Operator("sub", [self, other])
        return self.__add__(-other)

    def __rsub__(self, other):
        # if is_num(other) and other == 0:
        #     return -self
        # return Operator("sub", [other, self])
        return (-self).__radd__(other)
    
    # multiplication, puts the 'constant' (other) first
    def __mul__(self, other):
        if is_num(other) and other == 1:
            return self
        # this unnecessarily complicates wsum creation
        #if is_num(other) and other == 0:
        #    return other
        return Operator("mul", [self, other])

    def __rmul__(self, other):
        if is_num(other) and other == 1:
            return self
        # this unnecessarily complicates wsum creation
        #if is_num(other) and other == 0:
        #    return other
        return Operator("mul", [other, self])

    # matrix multipliciation TODO?
    #object.__matmul__(self, other)

    # other mathematical ones
    def __truediv__(self, other):
        warnings.warn("We only support floordivision, use // in stead of /", SyntaxWarning)
        return self.__floordiv__(other)

    def __rtruediv__(self, other):
        warnings.warn("We only support floordivision, use // in stead of /", SyntaxWarning)
        return self.__rfloordiv__(other)

    def __floordiv__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("div", [self, other])

    def __rfloordiv__(self, other):
        return Operator("div", [other, self])

    def __mod__(self, other):
        return Operator("mod", [self, other])

    def __rmod__(self, other):
        return Operator("mod", [other, self])

    def __pow__(self, other, modulo=None):
        assert (modulo is None), "Power operator: modulo not supported"
        if is_num(other):
            if other == 1:
                return self
        return Operator("pow", [self, other])

    def __rpow__(self, other, modulo=None):
        assert (modulo is None), "Power operator: modulo not supported"
        return Operator("pow", [other, self])

    # Not implemented: (yet?)
    #object.__divmod__(self, other)

    # unary mathematical operators
    def __neg__(self):
        # special case, -(w*x) -> -w*x
        if self.name == 'mul' and is_num(self.args[0]):
            return Operator(self.name, [-self.args[0], self.args[1]])
        elif self.name == 'wsum':
            # negate the constant weights
            return Operator(self.name, [[-a for a in self.args[0]], self.args[1]])
        return Operator("-", [self])

    def __pos__(self):
        return self

    def __abs__(self):
        return cp.Abs(self)

    def __invert__(self):
        if not (is_boolexpr(self)):
            raise TypeError("Not operator is only allowed on boolean expressions: {0}".format(self))
        return Operator("not", [self])

    def __bool__(self):
        raise ValueError(f"__bool__ should not be called on a CPMPy expression {self} as it will always return True\n"
                         "Do not use an expression as argument in an `if` statement and use cpmpy.any, cpmpy.max instead of python builtins\n"
                         "If you think this is an error, please report on github")

class BoolVal(Expression):
    """
        Wrapper for python or numpy BoolVals
    """

    def __init__(self, arg):
        assert is_true_cst(arg) or is_false_cst(arg), f"BoolVal must be initialized with a boolean constant, got {arg} of type {type(arg)}"
        super(BoolVal, self).__init__("boolval", [bool(arg)])

    def value(self):
        return self.args[0]

    def __invert__(self):
        return BoolVal(not self.args[0])

    def __bool__(self):
        """Called to implement truth value testing and the built-in operation bool(), return stored value"""
        return self.args[0]

    def __int__(self):
        """Called to implement conversion to numerical"""
        return int(self.args[0])

    def get_bounds(self):
        v = int(self.args[0])
        return (v,v)

    def __and__(self, other):
        if is_bool(other): # Boolean constant
            return BoolVal(self.args[0] and other)
        elif isinstance(other, Expression) and other.is_bool():
            if self.args[0]:
                return other
            else:
                return BoolVal(False)
        raise ValueError(f"{self}&{other} is not valid. Expected Boolean constant or Boolean Expression, but got {other} of type {type(other)}.")
        
    
    def __rand__(self, other):
        if is_bool(other): # Boolean constant
            return BoolVal(self.args[0] and other)
        elif isinstance(other, Expression) and other.is_bool():
            if self.args[0]:
                return other
            else:
                return BoolVal(False)
        raise ValueError(f"{self}&{other} is not valid. Expected Boolean constant or Boolean Expression, but got {other} of type {type(other)}.")

    
    def __or__(self, other):
        if is_bool(other): # Boolean constant
            return BoolVal(self.args[0] or other)
        elif isinstance(other, Expression) and other.is_bool():
            if not self.args[0]:
                return other
            else:
                return BoolVal(True)
        raise ValueError(f"{self}|{other} is not valid. Expected Boolean constant or Boolean Expression, but got {other} of type {type(other)}.")
        
        
    def __ror__(self, other):
        if is_bool(other): # Boolean constant
            return BoolVal(self.args[0] or other)
        elif isinstance(other, Expression) and other.is_bool():
            if not self.args[0]:
                return other
            else:
                return BoolVal(True)
        raise ValueError(f"{self}|{other} is not valid. Expected Boolean constant or Boolean Expression, but got {other} of type {type(other)}.")
        
    def __xor__(self, other):
        if is_bool(other): # Boolean constant
            return BoolVal(self.args[0] ^ other)
        elif isinstance(other, Expression) and other.is_bool():
            if self.args[0]:
                return ~other
            else:
                return other
        raise ValueError(f"{self}^^{other} is not valid. Expected Boolean constant or Boolean Expression, but got {other} of type {type(other)}.")
    
    
    def __rxor__(self, other):
        if is_bool(other): # Boolean constant
            return BoolVal(self.args[0] ^ other)
        elif isinstance(other, Expression) and other.is_bool():
            if self.args[0]:
                return ~other
            else:
                return other
        raise ValueError(f"{self}^^{other} is not valid. Expected Boolean constant or Boolean Expression, but got {other} of type {type(other)}.")
    

    def has_subexpr(self) -> bool:
        """ Does it contains nested Expressions (anything other than a _NumVarImpl or a constant)?
            Is of importance when deciding whether certain transformations are needed
            along particular paths of the expression tree.
        """
        return False # BoolVal is a wrapper for a python or numpy constant boolean.

    def implies(self, other):
        if self.args[0]:
            return other
        else:
            return other == other  # Always true, but keep variables in the model


class Comparison(Expression):
    """Represents a comparison between two sub-expressions
    """
    allowed = {'==', '!=', '<=', '<', '>=', '>'}

    def __init__(self, name, left, right):
        assert (name in Comparison.allowed), f"Symbol {name} not allowed"
        super().__init__(name, [left, right])

    def __repr__(self):
        if all(isinstance(x, Expression) for x in self.args):
            return "({}) {} ({})".format(self.args[0], self.name, self.args[1]) 
        # if not: prettier printing without braces
        return "{} {} {}".format(self.args[0], self.name, self.args[1]) 
    
    def __bool__(self):
        # will be called when comparing elements in a container, but always with `==`
        if self.name == "==":
            return repr(self.args[0]) == repr(self.args[1])
        super().__bool__() # default to exception

    # return the value of the expression
    # optional, default: None
    def value(self):
        arg_vals = argvals(self.args)

        if any(a is None for a in arg_vals): return None
        if   self.name == "==": return arg_vals[0] == arg_vals[1]
        elif self.name == "!=": return arg_vals[0] != arg_vals[1]
        elif self.name == "<":  return arg_vals[0] < arg_vals[1]
        elif self.name == "<=": return arg_vals[0] <= arg_vals[1]
        elif self.name == ">":  return arg_vals[0] > arg_vals[1]
        elif self.name == ">=": return arg_vals[0] >= arg_vals[1]
        return None # default


class Operator(Expression):
    """
    All kinds of mathematical and logical operators on expressions

    Convention for 2-ary operators: if one of the two is a constant,
    it is stored first (as expr[0]), this eases weighted sum detection
    """
    allowed = {
        #name: (arity, is_bool)       arity 0 = n-ary, min 2
        'and': (0, True),
        'or':  (0, True),
        '->':  (2, True),
        'not': (1, True),
        'sum': (0, False),
        'wsum': (2, False),
        'sub': (2, False), # x - y
        'mul': (2, False),
        'div': (2, False),
        'mod': (2, False),
        'pow': (2, False),
        '-':   (1, False), # -x
    }
    printmap = {'sum': '+', 'sub': '-', 'mul': '*', 'div': '//'}

    def __init__(self, name, arg_list):
        # sanity checks
        assert (name in Operator.allowed), "Operator {} not allowed".format(name)
        arity, is_bool_op = Operator.allowed[name]
        if is_bool_op:
            #only boolean arguments allowed
            for arg in arg_list:
                if not is_boolexpr(arg):
                    raise TypeError("{}-operator only accepts boolean arguments, not {}".format(name,arg))
        if arity == 0:
            arg_list = flatlist(arg_list)
            assert (len(arg_list) >= 1), "Operator: n-ary operators require at least one argument"
        else:
            assert (len(arg_list) == arity), "Operator: {}, number of arguments must be {}".format(name, arity)

        # automatic weighted sum (wsum) creation:
        # if all args are an expression (not a constant)
        #    and one of the args is a wsum,
        #                    or a product of a constant and an expression,
        # then create a wsum of weights,expressions over all
        if name == 'sum' and \
                all(not is_num(a) for a in arg_list) and \
                any(_wsum_should(a) for a in arg_list):
            we = [_wsum_make(a) for a in arg_list]
            w = [wi for w, _ in we for wi in w]
            e = [ei for _, e in we for ei in e]
            name = 'wsum'
            arg_list = [w, e]

        # we have the requirement that weighted sums are [weights, expressions]
        if name == 'wsum':
            assert all(is_num(a) for a in arg_list[0]), "wsum: arg0 has to be all constants but is: "+str(arg_list[0])
            weights = []
            for w in arg_list[0]:
                if is_int(w):
                    weights.append(int(w)) # bool or int, simplifies things later on
                else:
                    weights.append(w) # can be float
            arg_list = (weights, arg_list[1])

        # small cleanup: nested n-ary operators are merged into the toplevel
        # (this is actually against our design principle of creating
        #  expressions the way the user wrote them)
        if arity == 0:
            i = 0 # length can change
            while i < len(arg_list):
                if isinstance(arg_list[i], Operator) and arg_list[i].name == name:
                    # merge args in at this position
                    l = len(arg_list[i].args)
                    arg_list[i:i+1] = arg_list[i].args
                    i += l
                i += 1

        # another cleanup, translate -(v*c) to v*-c
        if hasattr(arg_list[0], 'name'):
            if name == '-' and arg_list[0].name == 'mul' and len(arg_list[0].args) == 2:
                mul_args = arg_list[0].args
                if is_num(mul_args[0]):
                    name = 'mul'
                    arg_list = (-mul_args[0], mul_args[1])
                elif is_num(mul_args[1]):
                    name = 'mul'
                    arg_list = (mul_args[0], -mul_args[1])

        super().__init__(name, arg_list)

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return Operator.allowed[self.name][1]

    def __repr__(self):
        printname = self.name
        if printname in Operator.printmap:
            printname = Operator.printmap[printname]

        # special cases
        if self.name == '-': # unary -
            return "-({})".format(self.args[0])

        # weighted sum
        if self.name == 'wsum':
            return f"sum({self.args[0]} * {self.args[1]})"

        # infix printing of two arguments
        if len(self.args) == 2:
            # bracketed printing of non-constants
            def wrap_bracket(arg):
                if isinstance(arg, Expression):
                    return f"({arg})"
                return arg
            return "{} {} {}".format(wrap_bracket(self.args[0]),
                                     printname,
                                     wrap_bracket(self.args[1]))

        return "{}({})".format(self.name, self.args)

    def value(self):

        if self.name == "wsum":
            # wsum: arg0 is list of constants, no .value() use as is
            arg_vals = [self.args[0], argvals(self.args[1])]
        else:
            arg_vals = argvals(self.args)


        if any(a is None for a in arg_vals): return None
        # non-boolean
        elif self.name == "sum": return sum(arg_vals)
        elif self.name == "wsum":
            val = np.dot(arg_vals[0], arg_vals[1]).item()
            if round(val) == val: # it is an integer
                return int(val)
            return val # can be a float
        elif self.name == "mul": return arg_vals[0] * arg_vals[1]
        elif self.name == "sub": return arg_vals[0] - arg_vals[1]
        elif self.name == "pow": return arg_vals[0] ** arg_vals[1]
        elif self.name == "-":   return -arg_vals[0]
        elif self.name == "div":
            try:
                return int(arg_vals[0] / arg_vals[1])  # integer division
            except ZeroDivisionError:
                raise IncompleteFunctionError(f"Division by zero during value computation for expression {self}"
                                              + "\n Use argval(expr) to get the value of expr with relational "
                                                "semantics.")
        elif self.name == "mod":
            try:# modulo defined with integer division
                return arg_vals[0] - arg_vals[1] * int(arg_vals[0] / arg_vals[1])
            except ZeroDivisionError:
                raise IncompleteFunctionError(f"Division by zero during value computation for expression {self}"
                                              + "\n Use argval(expr) to get the value of expr with relational "
                                                "semantics.")


        # boolean
        elif self.name == "and": return all(arg_vals)
        elif self.name == "or" : return any(arg_vals)
        elif self.name == "->": return (not arg_vals[0]) or arg_vals[1]
        elif self.name == "not": return not arg_vals[0]

        return None # default

    def get_bounds(self):
        """
        Returns an estimate of lower and upper bound of the expression.
        These bounds are safe: all possible values for the expression agree with the bounds.
        These bounds are not tight: it may be possible that the bound itself is not a possible value for the expression.
        """
        if self.is_bool():
            return 0, 1 #boolean
        elif self.name == 'mul':
            lb1, ub1 = get_bounds(self.args[0])
            lb2, ub2 = get_bounds(self.args[1])
            bounds = [lb1 * lb2, lb1 * ub2, ub1 * lb2, ub1 * ub2]
            lowerbound, upperbound = min(bounds), max(bounds)
        elif self.name == 'sum':
            lbs, ubs = get_bounds(self.args)
            lowerbound, upperbound = sum(lbs), sum(ubs)
        elif self.name == 'wsum':
            weights, vars = self.args
            bounds = []
            lowerbound, upperbound = 0,0
            #this may seem like too many lines, but avoiding np.sum avoids overflowing things at int32 bounds
            for w, (lb, ub) in zip(weights, [get_bounds(arg) for arg in vars]):
                x,y = int(w) * lb, int(w) * ub
                if x <= y: # x is the lb of this arg
                    lowerbound += x
                    upperbound += y
                else:
                    lowerbound += y
                    upperbound += x

        elif self.name == 'sub':
            lb1, ub1 = get_bounds(self.args[0])
            lb2, ub2 = get_bounds(self.args[1])
            lowerbound, upperbound = lb1-ub2, ub1-lb2
        elif self.name == 'div':
            lb1, ub1 = get_bounds(self.args[0])
            lb2, ub2 = get_bounds(self.args[1])
            if lb2 <= 0 <= ub2:
                if lb2 == ub2:
                    raise ZeroDivisionError(f"Domain of {self.args[1]} only contains 0")
                if lb2 == 0:
                    lb2 = 1
                if ub2 == 0:
                    ub2 = -1
                bounds = [
                    int(lb1 / lb2), int(lb1 / -1), int(lb1 / 1), int(lb1 / ub2),
                    int(ub1 / lb2), int(ub1 / -1), int(ub1 / 1), int(ub1 / ub2)
                ]
            else:
                bounds = [int(lb1 / lb2), int(lb1 / ub2), int(ub1 / lb2), int(ub1 / ub2)]
            lowerbound, upperbound = min(bounds), max(bounds)
        elif self.name == 'mod':
            lb1, ub1 = get_bounds(self.args[0])
            lb2, ub2 = get_bounds(self.args[1])
            if lb2 == ub2 == 0:
                raise ZeroDivisionError("Domain of {} only contains 0".format(self.args[1]))
            # the (abs of) the maximum value of the remainder is always one smaller than the absolute value of the divisor
            lb = lb2 + (lb2 <= 0) - (lb2 >= 0)
            ub = ub2 + (ub2 <= 0) - (ub2 >= 0)
            if lb1 >= 0:  # result will be positive if first argument is positive
                return 0, max(-lb, ub, 0)  # lb = 0
            elif ub1 <= 0:  # result will be negative if first argument is negative
                return min(-ub, lb, 0), 0  # ub = 0
            return min(-ub, lb, 0), max(-lb, ub, 0)  # 0 should always be in the domain
        elif self.name == 'pow':
            lb1, ub1 = get_bounds(self.args[0])
            lb2, ub2 = get_bounds(self.args[1])
            if lb2 < 0:
                raise NotImplementedError(f"Power operator: For integer values, exponent must be non-negative: {self}")
            bounds = [lb1**lb2, lb1**ub2, ub1**lb2, ub1**ub2]
            if lb1 < 0 and 0 < ub2:  
                # The lower and upper bounds depend on either the largest or the second largest exponent 
                # value when the base term can be negative. 
                # E.g., (-2)^2 is positive, but (-2)^1 is negative, so for (-2)^[0,2] we also need to add (-2)^1.
                bounds += [lb1 ** (ub2 - 1), ub1 ** (ub2 - 1)] 
                # This approach is safe but not tight (e.g., [-2,-1]^2 will give (-2,4) as range instead of [1,4]).
            lowerbound, upperbound = min(bounds), max(bounds)

        elif self.name == '-':
            lb1, ub1 = get_bounds(self.args[0])
            lowerbound, upperbound = -ub1, -lb1

        if lowerbound == None:
            raise ValueError(f"Bound requested for unknown expression {self}, please report bug on github")
        if lowerbound > upperbound:
            #overflow happened
            raise OverflowError(f'Overflow when calculating bounds, your expression exceeds integer bounds: {self}')
        return lowerbound, upperbound
def _wsum_should(arg):
    """ Internal helper: should the arg be in a wsum instead of sum

    True if the arg is already a wsum,
    or if it is a product of a constant and an expression 
    (negation '-' does not mean it SHOULD be a wsum, because then
     all substractions are transformed into less readable wsums)
    """
    return isinstance(arg, Operator) and \
           (arg.name == 'wsum' or \
            (arg.name == 'mul' and len(arg.args) == 2 and \
             any(is_num(a) for a in arg.args)
            ) )

def _wsum_make(arg):
    """ Internal helper: prep the arg for wsum

    returns ([weights], [expressions]) where 'weights' are constants
    call only if arg is Operator
    """
    if arg.name == 'wsum':
        return arg.args
    elif arg.name == 'sum':
        return [1]*len(arg.args), arg.args
    elif arg.name == 'mul':
        if is_num(arg.args[0]):
            return [arg.args[0]], [arg.args[1]]
        elif is_num(arg.args[1]):
            return [arg.args[1]], [arg.args[0]]
        # else falls through to default below
    elif arg.name == '-':
        return [-1], [arg.args[0]]
    # default
    return [1], [arg]
